package sir.estimate;

//Import stuff
import java.io.*;
import java.util.*;
import org.omegahat.Simulation.RandomGenerators.*;
import org.omegahat.Probability.Distributions.*;
import org.omegahat.Simulation.MCMC.*;
import org.omegahat.Simulation.MCMC.Listeners.*;
import java.text.NumberFormat;
import java.text.DecimalFormat;


//Get the utility stuff.
import sir.util.*;

//Get the optimizer classes
import sir.optim.*;

//Get thet model
import sir.model.*;

/**
   This class implements estimation of the parameters in a SIR
   * epidemic. Four methods are supported:
   *
   * <ol>
   * <li> MCMC analysis as described in O'Neill & Roberts, JRSSA 162. The results
   * of MCMC using burn-in rate, thinning factor, etc. specified by the
   * code are written to a file log.txt. This can then be used by
   * CODA or BOA for analysis in Splus.
   * <li> Maximum Likelihood
   * <li> Maximum Marginalized Likelihood
   * <li> Simulated ML
   * </ol>
   *
   * The class design unfortunately shows clea signs that this has been an incremental
   * project, i.e. it has grown from nothing to something quite huge, without the structure
   * following the same development. E.g. this means that global variables and switches are used
   * instead of nice design patterns.
   *
   <ul>
   <li>Integrate SIM part into the program and change name to something more useful
   </ul>
   @author Michael Hhle &lt;<a href="mailto://hoehle@dina.kvl.dk">hoehle@dina.kvl.dk</a>&gt;
   @version 2 April 2002
   */

public class LadyBug extends BaseMarkovChain {
  /** Debugging flag - gives extra printed info */
  public static boolean DEBUG= false; //false;// = true; //

  //True mass models are currently not implemented.
  ///** Population size, in true-mass it is N+a otherwise 1 */
  //protected double  popSize;

  /** Do we have acces to I time */    public static boolean gotI;

   /** New version - all data related variables are gathered in a Data object */
   protected Data data = null;

   /** Variable to keep track of acceptance rate. This is only relevant for MCMC analysis.
    *  Why not make a class for all the stuff related to MCMC instead? */
   //protected int runs = 0, accepted = 0, betaaccept = 0, betaNaccept = 0;

   /** Do we want the beta's to be updated using Metropolis sampling. Usually the answer is
    *  no if data.noOfUnits == 1 and yes otherwise. But it is possible to hardcode it.
    */
   protected boolean metropolisBeta = false;
   /** Do we want the gamma's to be updated using Metropolis, i.e. RandomWalk */
   protected boolean metropolisGamma = false;

   /** For formatting of doubles to an output stream. */
   //public static NumberFormat form;
   public static DecimalFormat form;

   /** The Pseudo random generator for distributions */
   public static Distributions prob;

   /** Configuration settings */
   public static boolean E0Fixed = false;
//   public static boolean IFixed = false;
   /** Should the betaN values be fixed or are we using a Metropolis sampler here? */
   boolean noBetaNUpdate = false;
   boolean thetaUpdate = false;

   /** Model to use in the estimation */
   EpiState model;

   /** Generate the next state of the Markov Chain using
    * Gibbs-within-Metropolis. In this new version all the business
    * is done by the state.
    *
     @param state Current chain of the MC
     @return New state of chain.
     @see EpiState
     */
   public MCMCState generate( MCMCState state ) {
     if (state == null) {
       state = model;
       return(((EpiState)state).sampleState(false));
     } else {
       return(((EpiState)state).generateNext());
     }
   }

   /**
       Read information about the epidemic from the denoted files.
       This is a new version, where inf and rec times are in the same file.
       The JavaCC file DataFileSpec.jj contains the specification.

       @param fileName The filename of the data file.
       */

   protected void readDataFile( String fileName) {
     String aLine,header;

     System.out.println("---------- Reading data ----------\n");
     //////////////////////////////////////////////////////////////////
     //
     // Read the new combined data file
     //
     /////////////////////////////////////////////////////////////////
     //Open the file for reading
     try {
       //Create parser and associate file to it.
       DataFileSpec parser = new DataFileSpec(new FileInputStream(fileName));

       try {
         //Do the parsing, i.e. start with first production.
         parser.DataFile();
         data = parser.getData();

         //Update Datalink in SIRState - NOT VERY NICE
         SEIRState.setData(data);
         SEIRState.setProb(prob);

         //Make sure our data is valid.
         if (!data.isOk()) {
           System.exit(0);
         }
         System.out.println(data);
       } catch (ParseException e) {
         System.err.println("Error: Problems while parsing " + fileName + ".");
         System.out.println(e);
         System.exit(0);
       }
     } catch (java.io.FileNotFoundException e) {
       System.out.println(e);
       System.exit(0);
     }
   }


   /** Constructor which calls all the necessary sub-routines,
    * s.a. reading of data, registering of listeners and iteration of
    * the MC. Output is written into the file log.txt.
    *
    * @param prng        The pseudo random number generator
    * @param dataFile  Name of the file containing recovery, inf and sero times
    */
   public LadyBug(PRNG prng, String dataFile) {
     //Initialize
     prob = new Distributions(prng);

     //Setup the formatting to US style and 5 digits
     /*form = NumberFormat.getInstance(Locale.ENGLISH);
     form.setMinimumFractionDigits(8); //4
     form.setMaximumFractionDigits(8); //4
     */
     Locale.setDefault(Locale.ENGLISH);
     //form = new DecimalFormat("####.000000");
     form = new DecimalFormat("0.000E00");
     //form = new DecimalFormat("0.0000");

     //Read data & initialize based on data
     readDataFile(dataFile);
   }

   /**
    * MCMC Estimation method - generates noOfSamples*thinFactor + noOfBurnin samples
    * and estimates beta and gamma based on recovery times alone.
    *
    * @param noOfSamples The size of the sample to generate
    * @param noOfBurnin Number of samples to use as burnin. Note: No thinning applied here
    * @param thinFactor Thinning, e.g. take each 100th sample.
    * @param logFileName Name of the log file to save the output.
    */
   public void MCMCEst(int noOfSamples, int thinFactor, int noOfBurnin,String logFileName) {
     //Decide whether to have the additional diagnostic information or to do without it.
     //I.e. if incuTimePDF = none or no diagnostic times are parametric
     if (data.getNoOfUnknownD() == data.noOfIndividuals) {
       model = new SEIRState();
     } else {
       model = new SEIRwDState();
     }

     //Listener
     MyStrippedListenerWriter listen = null;
     try {
       //Register burnin and thinning listeners.
       BurninProxyListener burnin = new BurninProxyListener(noOfBurnin);
       this.registerListener(burnin);

       ThinningProxyListener thinner = new ThinningProxyListener(thinFactor);
       burnin.registerListener(thinner);

       listen = new MyStrippedListenerWriter(logFileName);
       thinner.registerListener(listen);

       //Create the header information to write to the stream.
       listen.writeHeader(model.toHeader());
     } catch (java.io.IOException e) {
       System.out.println(e);
     }
     //Compute desired number of iterations:
     long noOfIterations = noOfSamples*thinFactor + noOfBurnin;
     System.out.println("Number of desired iterations: " + noOfIterations );

     //Do the iterations.
     iterate( (int)noOfIterations );
     //Close the log file.
     listen.close();

     /////////////////////////////////////////////////////////////////////////
     //Show acceptance rate statistics.
     ////////////////////////////////////////////////////////////////////////
     System.out.println(model.showStats(logFileName));
   }

   /**
    * Calculate parameters beta and gamma using ML estimation.
    *
    * @param printRes If true, the parameter estimates found are printed to System.out.
    * @param state In case we want to supply infection times directly, we provide them. If null then they are taken from the I-file.
    * @return SIRState object containing the parameter estimates.
    */
   public void MLEst() {
     SEIRState state = null;
     //Check we got full data.
     if ((data.getNoOfUnknownE() + data.getNoOfUnknownI()+data.getNoOfUnknownR())  != 0) {
       System.err.println("Error: To do ML Estimation, the epidemic has to be fully observed.");
       System.exit(-1);
     } else {
       //If all seroconversions are observed we are using a SEIRwD model
       if ((data.getNoOfUnknownD()-data.getNoOfCensoredD()) == 0) {
         state = new SEIRwDState();
       } else {
         //Otherwise, all seroconversion times are ignored and we use a SEIR model
         state = new SEIRState();
       }
     }
     double foo[][] = new double[state.noOfModelParams+1][state.noOfModelParams+1];
     state.ML(foo);
     //To create stuff for report figures.
     //state.MLProfiles();
   }

   /**
    * Main function - start a LadyBug analyserer based on prompt arguments
    * @param argc 1st argument = text file wit SIR information.
    * @see LadyBug#readDataFile
    */
   static public void main( String[] argv ) throws Throwable {
     //Show splash screen
     System.out.println(
        "\n-=-=-=-=-=-=-=-=-=-=-=--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-");
     System.out.println(
        "-      \\./     -=-=  LadyBug v2.0 by Michael Hoehle              =-=-=-=-=");
     System.out.println(
        "-     (o o)    -=-=  http://www.stat.uni-muenchen.de/~hoehle/    =-=-=-=-=");
     System.out.println(
        "=-oOOo-(_)-oOOo-=-=-=-=-=--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=");
     System.out.println("-                                                                        -");
     System.out.println("- LadyBug 2.0 comes with ABSOLUTELY NO WARRANTY; for details see         -");
     System.out.println("- http://www.gnu.org/copyleft/gpl.html (GPL). This is free software,     -");
     System.out.println("- and you are welcome to redistribute it under certain conditions        -");
     System.out.println("- as specified by GPL.                                                   -"); 
     System.out.println("-                                                                        -");
     System.out.println("=-=-=-=-=-=-=-=-=-=-=-=-=--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=\n");


     //Command agruments order
     int dataFileNo = 0;
     int optionsFileNo = 1;
     int logFileNo = 2;
     
     try {
       //Open the Options file for reading
       try {
         //Create parser and associate file to it.
         OptionsFileSpec parser = new OptionsFileSpec(new FileInputStream(argv[
             optionsFileNo]));
         
         try {
           //Do the parsing.
           parser.OptionsFile();
           //Set the Options as being read from file.
           Options.readFromFile = true;
           //Re-init-parser
           //InputStream foo = null;
           //parser.ReInit(foo);
         }
         catch (ParseException e) {
           System.err.println("Error: Problems while parsing " +
                              argv[optionsFileNo] + ".");
           System.out.println(e);
           System.exit(0);
         }
       }
       catch (java.io.FileNotFoundException e) {
         System.out.println(e);
         System.exit(0);
       }
       if (DEBUG) {
         System.out.println(new Options().toString());
       }
       ////////////////////////////////////////////////////////////////
       // Check whether seed for random generator should be "random"
       // or a fixed value so we can replicate results.
       ////////////////////////////////////////////////////////////////
       int seed;
       if (Options.fixedSeed) {
         //Fixed seeding - in order to be able to restore results.
         seed = Options.seed;
       }
       else {
         //Use Java's own random generator to seed our randomgen.
         java.util.Random r = new java.util.Random();
         seed = Math.abs(r.nextInt()); //should be 1 - 2^31-1 and notjust abs
       }
       //Now create the PRN generator
       CollingsPRNGAdministrator a = new CollingsPRNGAdministrator(seed);
       PRNG prng = new CollingsPRNG(a.registerPRNGState());
    
       //Create a SIR object, which reads its data from file and initialized
       //depending on whether its a true mass model or not.
       LadyBug sir = new LadyBug(prng, argv[dataFileNo]);
       
       //Start the timer
       long start = System.currentTimeMillis();
 
       //Big switch to find out what type of estimation routine to do.
       //Might be better style to have an Estimation interface each routine
       //implementing their own run method.
       System.out.println(
          "-------------------------------\nSelected estimation method: " +
          Options.method + "\n");
       if (Options.method.equals("ml")) {
         sir.MLEst();
       }
       else if (Options.method.equals("mcmc")) {
         //MCMC Estimation requires that m>1
         if (sir.data.noOfIndividuals > 1) {
           sir.MCMCEst(Options.mcmc_samples, Options.mcmc_thin,
                       Options.mcmc_burnin,argv[logFileNo]);
         }
         else {
           System.err.println("Error: To do MCMC estimation observed number of recoveries has to be greater than 1!");
         }
       } 
       System.out.println("\nJob done...used " +
                          ( (System.currentTimeMillis() - start) / 1000) + " s.");
     } catch (ArrayIndexOutOfBoundsException e) {
       //If the no of parameters is wrong then show the correct syntax
       System.err.println(e + "\n");
       System.err.println("Syntax Error:\n" +  "LadyBug <dataFileName> <optionsFileName> [logFileName (MCMC)]");
       
     }
   }
 }
       
       

