package sir.model;

import java.text.NumberFormat;
import java.io.*;
import java.util.*;
import org.omegahat.Simulation.MCMC.*;

//My own sbetuff.
import sir.optim.*;
import sir.util.*;
import sir.estimate.*;

  /**
   * The EpiState is an abstract base class implementing common features of the epidemic
   * model.
   * 
   * A class hierarchy with
   * SEIRState and SEIRwDState as subclasses. The EpiState is an abstract base class implementing
   * a couple of common features -- most functionality will be made in the subclasses.
   *
   * 
   * @author Michael Hhle &lt;hoehle@stat.uni-muenchen.de&gt;
   * @version 16 October 2002
   *
  */
  public abstract class EpiState implements MCMCState,Uncmin_methods {
	/** Array of infection times, I1,...,Im */
	public Event[] I;
       	/** Array of recovery times, R1,...,Rm */
	public Event[] R;
        /** Number of model parameters, i.e. no of unknowns when we have full data */
        public int noOfModelParams;
        /** The names of all model parameters to be used when printing them */
        public String[] modelParamNames;
        /** To format numbers in layout. */
        NumberFormat form;

        /** Infection pressure aka. transmission parameter */
        public double beta;
        /** Neighbour effect */
        public double eta;
        /** Vaccination effect */
        public double betaN;
        /** Dual representation of params by array */
        public double theta[];

        //Index of model params in the parameter vector.
        public int beta_,betaN_;

        /** Array of indexes showing which infection times are unknown and hence are variables */
        int IParamIndex[];
        /** Number of unknown infection times */
        int noOfIParam;
        /** Array of indexes showing which recovery times are unknown and hence are variables */
        int RParamIndex[];
        /** Number of unknown recovery times */
        int noOfRParam;

        /** A list containing all parametric events of the state. */
        LinkedList paramList = new LinkedList();

        /** LogLikelihood */
        public double loglik = 0;
        /** Boolean telling whether the time wise structure has been created on this object (inits events array)*/
        public boolean nogen = true;
        /** Array containing events */
        Event events[] = null;

        /** The number of events */
        int noOfEvents;

        /** Static stuff to keep track of progress */
        public static int runs = 0;

        /** Acceptance counts for all traditional parameters*/
        public int[] accept;
        //And two traditional ones - grouped.
        double Iaccept;
        double Raccept;

        /** Static link to the data. */
        protected static Data data;
        /** Static link to a random number generator */
        protected static Distributions prob;

        /** Take a line from a logfile and assign the parameters. Caution: We assumed the logfile
         * has correct format compared to the instance of EpiState we are looking at.
         * 
         * @param aLine  A line from the logfile containing all the parameters
         */
        public void fromString(String aLine) {
          //Tokenize it with usual whitespace (\t the only relevant). 
          StringTokenizer st = new StringTokenizer(aLine);
          //Convert to something useful
          SEIRState ss = ((SEIRState)this);
           
          //Be prepared for parse exceptions.
          try {
            //LogLik
            loglik = form.parse(st.nextToken()).doubleValue();
            
            //Read all ordinary parameters (theta)
            for (int i=1; i<=this.noOfModelParams; i++) {
              theta[i] = form.parse(st.nextToken()).doubleValue();
            }
            //Convert theta to labeled parameters.
            ss.array2Params(theta,false);
           
            //Exposure times - actually length of incubation is shown, i.e. T_E. So 
            //we read their values now and re-calculate once the I times are known.
             for (int i=0; i<data.noOfIndividuals; i++) {
               ss.E[i].t = form.parse(st.nextToken()).doubleValue();
             }


            //Infection Times
            for (int i=0; i<noOfIParam; i++) {
              ss.I[IParamIndex[i]].t =  form.parse(st.nextToken()).doubleValue(); 
            }
           
             //Recalculate exposure times.
             for (int i=0; i<data.noOfIndividuals; i++) {
               ss.E[i].t = ss.I[i].t-ss.E[i].t;
             }



            //Diagnostic times(should there bere some kind of switch here==!?!"!"
            SEIRwDState sswd = ((SEIRwDState)this);
            for (int i=0; i<sswd.noOfDParam; i++) {
              sswd.D[sswd.DParamIndex[i]].t = form.parse(st.nextToken()).doubleValue();
            }
          } catch (java.text.ParseException e) {
            System.err.println(e);
          }
          //Sort, order, whatever it takes to get order!
           genXYFuncs();
          //System.out.println(toStringLong());
          //System.out.println(((SEIRwDState)this).());
        }
        
        //Empty hook function.
        public String toStringHook() { return("");}
        
          /** Debug/Information as usual. This is used in logfiles */
	public String toString() {
	    StringBuffer s = new StringBuffer();
            //Log likelihood is not of interest atm, but let's compute it anyway...?
            //Would be less time consuming to have it stored in a variable.
            //generateNext updates this value everytime..
            s.append(form.format(loglik));

            //Put all parameter values into string
            double theta[] = params2Array();
            for (int i=1; i<=this.noOfModelParams; i++) {
              s.append("\t" + form.format(theta[i]));
            }

            s.append(toStringHook());
            
           
            //Return result
            return(s.toString());
        }

        /** Current State information is shown on screen in long format*/
        public String toStringLong() {
          return("");
        }

        /** What to write to the header of the log*/
        public String toHeader() {
          StringBuffer s = new StringBuffer();
          s.append("loglik");
          //Put parameters into the string
          for (int i=1;i<=noOfModelParams;i++) {
            s.append("\t" + modelParamNames[i]);
          }
          for (int i=0; i<data.noOfIndividuals; i++) {
            s.append("\t" + "te" + i);
          }
          for (int i=0; i<noOfIParam; i++) {
            s.append("\ti"+IParamIndex[i]);
          }
          
          //Add additional stuff depending on model class.
          s.append(toHeaderHook());
         
          return(s.toString());
        }
        
        //EMpty hook
        public String toHeaderHook() { return("");}
        
        /** Show X(t) and Y(t) as functions */
        public String toStringEvents() {
          StringBuffer s = new StringBuffer();
          s.append("\n");
          for (int i=0; i<noOfEvents; i++) {
            Event e = events[i];
            s.append(e.toString() +"\n");
            //s.append(LadyBug.form.format(e.t) + "\t" + e.unit.posx + "\t" + e.unit.posy + "\t" +  e.type + "\t" + e.X + "\t" + e.Y + "\n");
          }
          return(s.toString());
        }
        /**
         * Calculate int_{i_0}^t Y_u(t) dt for a unit u and regime r.
         *
         * @param u The unit to get Y(t) from.
         * @param tstop Integrate \int_{i_0}^{t} Y(t) dt
         * @return The integral.
        */
        protected double Yintegrate(Unit u, double tstop) {
          double res = 0;

          //In case we are interested to stop before I_0, we are done.
          if (tstop < u.firstEvent.t) {return(0); }

          //Integrate as far as we can get.
          Event elast = data.units[u.posx][u.posy].firstEvent;
          Event e = elast.nextEventInSameUnit;
          while ((e != null) && (e.t < tstop)) {
            //System.out.println("Added " + elast.Y + " * " + e.t + "\t-\t" +  elast.t);
            res += elast.Y * (e.t - elast.t);
            elast = e;
            e = e.nextEventInSameUnit;
          }
          //Add the rest, which was lying between events.
          res += elast.Y * (tstop-elast.t);
          //System.out.println("Extra Added " + elast.Y + " * " + tstop + "\t-\t" +  elast.t);
          //System.out.println("The result of Yintegrate is " + res);
          return(res);
        }

        /** Hook function to be used by subclasses */
        public  boolean isValidHook() {return true;}

       /**
         Verify whether I is valid wrt. tau, i.e. that the constraints are fulfilled
       */
        public boolean isIValid() {
          //Create necessary constructions
          if (this.nogen) {genXYFuncs();}

          //If X,Y,Z are NAN we definitely have a problem and do not need to visit the likelihood
          if (Double.isNaN(events[0].X)) {return(false);}

          //If it's a transmission experiment no negative event times are allowed
          if (!isValidHook()) {return(false);}

          //Calculate likelihood
          Double logL = new Double(loglik());
          System.out.println("log lik=" + logL);
          return(!logL.isInfinite() && !logL.isNaN());
       }

      /** Generate X, Y as functions of time by bookeeping of events */
      public abstract void genXYFuncs();

      /** Abstract method to calculate the loglikelihood of the state
       */
      public abstract double loglik();

      /** Make a valid instance of the parameters, e.g. to start the MCMC chain with
       *
       * @param authentic Boolean telling whether to sample a value for censored events.
       *                  This is typical in MCMC but not in MLE.
       */
      public abstract EpiState sampleState(boolean sampleCensored);

      /** Based on acceptance probability we find out whether to accept a sample
       * i.e. we accept it with probability min( f(x'|)pi(x') / f(x|)pi(x), 1)
       *
       * @param  logfnew log( f(x'|...) * pi(x') ), i.e. numerator of acceptance fraction
       * @param  logfold log( f(x'|...) * pi(x)  ), i.e. denumerator of acceptance fraction
       * @return Boolean telling whether the proposal is to be accepted.
       */
      protected boolean acceptProposal(double logfnew, double logfold) {
        double acceptProb = Math.min(Math.exp(logfnew-logfold),1);
        double uu = prob.getPRNG().nextDouble();

        if (LadyBug.DEBUG) {
          System.out.println("logfnew = " + logfnew);
          System.out.println("logfold = " + logfold);
          System.out.println("acceptprob = " + acceptProb + " u = " + uu);
        }

        //Accept or reject the new value?
        return(!((Double.isNaN(acceptProb)) || (uu >= acceptProb)));
      }


      /** Set reference to the data */
      public static void setData(Data theData) { data = theData; }
      public static void setProb(Distributions theProb) { prob = theProb; }

      /** Generate method for MCMC chain 
       * @param justTheta Boolean indicating whether to just update the beta,...,gamma_D variables
       * 
       */
      public abstract EpiState generateNext(boolean justTheta);

      public EpiState generateNext() { return(generateNext(false)); } 
      /** Helper functions to put all parameters in an array. Remember Fortran indexing
       * Function currently does not work if incu period a constant.
       *
       * @return An @see noOfModelParams + 1 long array with params
       */
      public abstract double[] params2Array();


      /**
       * Method to present mean and variance of posterior for each variable contained in the logfile.
       * This is a new generic version handling any amount of parameters
       *
       * @param logFileName The name of the logfile to look in.
       * @return String with the information to present
       */
      public String analLog(String logFileName) {
        StringBuffer s = new StringBuffer();

        //Init the places to store all the logfile values. Make array one longer coz we
        //use Fortran indexing of params
        Vector params[] = new Vector[noOfModelParams+1];
        for (int i=1; i<=noOfModelParams; i++) { params[i] = new Vector(); }

        try {
          FileInputStream fin =  new FileInputStream(logFileName);
          BufferedReader buf  = new BufferedReader(new InputStreamReader(fin));

          //Skip header
          String header = buf.readLine();
          String aLine;

          //Start reading line by line
          while (((aLine = buf.readLine()) != null)) {
            StringTokenizer st = new StringTokenizer(aLine);
            //Fetch loglik value
            double loglik = form.parse(st.nextToken()).doubleValue();

            //Read all the parameters from the file
            for (int i=1; i<=noOfModelParams; i++) {
              params[i].add(new Double(form.parse(st.nextToken()).doubleValue()));
            }
          } // end of reading from logfile

          //Calculate means
          double means[] = new double[noOfModelParams+1];
          for (int i=1; i<=noOfModelParams; i++) {
            for (int j=0; j<params[i].size(); j++) {
              means[i] += ((Double)params[i].elementAt(j)).doubleValue()/params[i].size();
            }
          }

          //Calculate variance
          double vars[] = new double[noOfModelParams+1];
          for (int i=1; i<=noOfModelParams; i++) {
            for (int j=0; j<params[i].size(); j++) {
              vars[i] += Math.pow(((Double)params[i].elementAt(j)).doubleValue()  - means[i], 2)  /(params[i].size()-1);
            }
          }

          //Write the information to screen
          s.append("------ Posterior distributions (E,Sqrt(V)) -------\n");
          for (int i=1; i<=noOfModelParams; i++) {
            s.append(modelParamNames[i] + "\t= (" + LadyBug.form.format(means[i]) + " , " + LadyBug.form.format(Math.sqrt(vars[i]))+")\n");
          }

          //Shut down
          buf.close();
          fin.close();
        } catch(Exception e) {
          System.out.println(e);
        }
        return(s.toString());
      }
      /**
       * Show statistics once we are done.
       *
       * @param logFile Name of the logFile to analyze
       * @return StringBuffer Contains all the stat information
       */
      public String showStats(String logFile) {
        StringBuffer s = new StringBuffer();
        //Open logfile and show stats
        s.append(analLog(logFile));

        //Return the value.
        return(s.toString());
      }
      
     
      
      /**
       * Constructor - initializes structures holding who is parameter and who is fixed
       * among the recovery and infection events.
       *
       */
      public EpiState() {
        //Init number formatter
        form = LadyBug.form;

        //Fetch interesting numbers from the Data class
        noOfIParam = data.getNoOfUnknownI();
        noOfRParam = data.getNoOfUnknownR();
        IParamIndex = data.getUnknownIIndexes();
        RParamIndex = data.getUnknownRIndexes();
      }


  } //End of class SIRState

