package sir.model;

import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.io.*;
import java.util.*;
import org.omegahat.Simulation.MCMC.*;

//My own stuff.
import sir.optim.*;
import sir.util.*;
import sir.estimate.*;

/**
 * Subclass implementing all functionality for an multivariate SEIR model.
 * 
 * This class is used to represent the state as a SEIR model as mentioned in the
 * O'Neill and Becker article "Inference for an epidemic when susceptibility varies", 
 * P.D. O'Neill and Niels G. Becker, Biostatistics 2(1), pp.99-108 (2001) and
 * the M. Hoehle, E. Jrgensen, and P.D. O'Neill article "Inference in disease
 * transmission experiments using stochastic epidemic models.
 * 
 * I.e. we have
 * <ul><li> Constant or Gamma distributed latency time,
 *     <li> a hazard of beta* X * Y for infection 
 *     <li> Gamma distributed infection time, i.e. f I\sim \gammaI(\gammaI,\deltaI)
 *     <li> Multivariate setup <b>(so far just tested up to dimension 2!!)</b>
 * </ul>
 * Note: Earlier versions of this program contained the possibility to specify fixed times
 * where the set of parameters switched due to control measures etc. All this functionality
 * is now removed -- consult the reference <a href="http://www.stat.uni-muenchen.de/~hoehle/pubs/dina102.pdf">"Estimating Parameters for Stochastic Epidemics" by
 * M. Hhle and E. Jrgensen</a>, Dina Research Report no. 102, 2002.<p>
 * 
 * Copyright: GPL <br>
 * @author Michael Hhle &lt;hoehle@stat.uni-muenchen.de&gt;
 * @version 16 October 2002
 *
 */
public class SEIRState extends EpiState implements MCMCState,Uncmin_methods {
  /** Parameters of the gamma distributed incbuation time */
  public double gammaE, deltaE;
  /** Parameters of the gamma distributed infectious time */
  public double gammaI, deltaI;
  /** Index of inc time parameters in param vector */
  public int gammaE_, deltaE_, gammaI_, deltaI_;

  /** Re-parameterization using mean and variance */
  public boolean MeanVarianceParameters;
  /** Just a foobar thing - only use diagnostic times in the likelihood */
  public boolean justDiagnostic = false;

  /** Are we operating with log transformed parameters in the loglik function.
   * This is necassry when calling the optimizer, but not good when computing the Fisher info matrix
   */
  public boolean expTransformParams;
  /** All the exposed states */
  public Event[] E;
  //How many of the E's are not known
  int noOfEParam;
  //Array containing the index of the E's not known
  int[] EParamIndex;
  //Acceptance rate when updating */
  public static double Eaccept = 0;

  /** Are we using exposure origins in the loglik calculations. Always true except in ML calculations */
  public boolean useOrigins = true;

  /** First event in all units - a data structure updated each timegenXYFuncs is called.
   *  The purpose is to speed up the lastEventInit() method.
   */
  public Event[][] firstEventInUnits;
  /** Global variable to speed up the lastEventInit method so we do not need to new each time*/
  public Event[][] lastEventInUnits;

  /** Show how state params relate to data */
  public String toStringDiagnostic() {
    StringBuffer s= new StringBuffer();


    //How many real params are there really
    int paramCorrection = (data.noOfUnits==1) ? -1 : 0;
    paramCorrection -= (Options.diagTime == 'E') ? 1 : 0;
    paramCorrection -= (Options.expInfTime) ? 1 : 0;
    paramCorrection -= (Options.incuTime == 'E') ? 1 : 0;
    paramCorrection -= (Options.incuTime == 'C') ? 2 : 0;

    //What is the log likelihood of the current value
    s.append("Loglik(state)               = " + loglik()+"\n");
    s.append("AIC = -2*loglik + 2*" + (noOfModelParams+paramCorrection) + "       = " + (-2*loglik()+2*(this.noOfModelParams+paramCorrection)) + "\n\n");

    //Calculate mean and variances.
    double EV[] = this.periodsEV();
    //Data just shown in 
    s.append("Data Incubation period      = (" + form.format(EV[0]) + "\t, " + form.format(EV[2]) + ")\n");
    s.append("Data Infection  period      = (" + form.format(EV[1]) + "\t, " + form.format(EV[3]) + ")\n");

    if (MeanVarianceParameters) {
      if (Options.incuTime != 'C') {
        s.append("Estimated Incubation period = (" + form.format(gammaE) +
                 "\t, " + form.format(deltaE) + ")\n");
      } else {
        s.append("Data Incubation period      = " + data.incu + "\n");
      }
      s.append("Estimated Infectious period = (" + form.format(gammaI) + "\t, " + form.format(deltaI)+ ")\n");
    } else {
       if (Options.incuTime != 'C') {
         s.append("Estimated Incubation period = (" +
                  form.format(gammaE / deltaE) + "\t, " +
                  form.format(gammaE / Math.pow(deltaE, 2)) + ")\n");
       } else {
          s.append("Data Incubation period      = " + data.incu + "\n");
       }
      s.append("Estimated Infectious period = (" + form.format(gammaI/deltaI) + "\t, " + form.format(gammaI/Math.pow(deltaI,2))+ ")\n");
    }

    return(s.toString());
  }

  /** Current State information is shown on screen in long format*/
  public String toStringLong() {
    StringBuffer s = new StringBuffer();

    //Translate all parameters and print them
    double theta[] = params2Array();
    for (int i=1; i<=this.noOfModelParams; i++) {
      s.append(this.modelParamNames[i] + "\t= " + form.format(theta[i]) + "\n");
    }

    //Add diagnostic information
    s.append(this.toStringDiagnostic());

    //Return all info
    return(s.toString());
  }



  /** Method to compute \lambda_e^j (t-) when the origin of the event is known.
   *
   * @param u Unit, where the individual is located.
   * @param o Unit, where the exposure is originating from (o is equal to u or one of its neighbours)
   * @param le Array with last event occured in each section. I.e. state just before to the event of interest.
   * @return \lambda_e^j(t-) = beta^I(u=o) beta_n^I(o\in N_4(u)) Y_o(t)  zero otherwise
   */
  public double betaHaz(Unit u, Unit o, Event[][] le) {
    //Is there a vaccination effect?
    //double x = (u.vacc ? vr : 1);

    if (u == o) {
      //System.out.println("Y[" + o.posx + "][" + o.posy + "].Y = " + le[o.posx][o.posy].Y + ");
      //System.out.println("le event to multiply Y with:\n" + le[o.posx][o.posy]);
      return(beta  * le[o.posx][o.posy].Y);
    } else if (u.isNeighbour(o)) {
      return(betaN * le[o.posx][o.posy].Y);
    } else {
      return(0);
    }
  }

  /** Method to compute \lambda_{all}^j(t-|...) for a specific individual (located in u_j). That is
   *  the overall hazard exerted on a single individual just prior to occurence of an event.
   * Corresponds to the setup, where no locations are given.
   *
   * @param u Unit, where the individual is located.
   * @param le Array with last event occured in each section. I.e. state previous to the event of interest.
   * @return \lambda_{all}^j(t-|...) = beta*Y_u(t) + \sum_{u'\in N_4(u)} beta_n * Y_u'(t-)
   */
  public double betaHazAll(Unit u, Event[][] le) {
    // beta* I_u(t-)
    double haz = betaHaz(u,u,le);
    //System.out.println("haz=" + haz);
    // + \sum_{u'\in N_4(u) beta_n * I_u'(t-)
    for (int ox=1; ox<=data.noOfUnitsX; ox++) {
      for (int oy=1; oy<=data.noOfUnitsY; oy++) {
        if (u.isNeighbour(data.units[ox][oy])) {
          haz += betaHaz(u,data.units[ox][oy],le);
        }
      }
    }
    
    //Return result.
    return(haz);
  }


  /** Prepare the last event data structure. This is an optimized implementation
   *  exploiting that the genXYFuncs created the desired array and that we
   *  reuse an array instead of newing something every time.
   *  Newing is expensive, but can be necessary at times.
   *
   * @param newArray Should we new a new instance. Necessary if we have nested calls to lastEventInit as in loglik
   * @return Array where each unit is set to its first event.
   */
  public Event[][] lastEventInit(boolean newArray) {
    //Dummy variable for result
    Event le[][];

    if (newArray) { //new
      le = new Event[data.noOfUnitsX+1][data.noOfUnitsY+1];
    } else { //use global variable
      le = lastEventInUnits;
    }

    //Load the global instance lastEventInUnits with the data
    //from firstEventInUnits. Have to ensure, that we never call
    //lastEvenInit twice within each other...
    for (int x=1; x<=data.noOfUnitsX; x++) {
      for (int y=1; y<=data.noOfUnitsY; y++) {
        le[x][y] = firstEventInUnits[x][y];
      }
    }
    return(le);
  }


  /** Compute log S_j(t), i.e. the survival of an individual
   *  located in unit u up to time t, when overall hazard exerted on the individual
   *  is \lambda_j(t).
   *
   * @param u The unit, where the individual is located.
   * @param estop Time up to which to integrate. Will only allow event times.
   * @return -\int_{0)^{t} \lambda_j(u) du (notice this is log S_j(t))
   */
  public double betaSurvLn(Unit u,double tstop) {
    //The result
    double surv = 0;

    //Init the last event structure
    Event le[][] = this.lastEventInit(false);
    //Init the current event e_t and the event just before it e_{t-1}
    int i=1;
    Event em1 = events[0];
    Event e = events[i];

    //In case we are interested to stop before I_0, we are done.
    if (tstop < events[0].t) {return(0);}

    //Integrate as far as we can get.
    while ((e != null) && (e.t < tstop)) {
      //Integrate the functions just up to the event occuring. I.e. we always take
      //the area under the curve to the left (time-wise) of the event.
      //System.out.println("Adding " + betaHazAll(u,le) + " for time interval " + e.t + " - "  + em1.t);
      surv += betaHazAll(u,le)*(e.t-em1.t);

      //Advance and keep track of the last event we had a look at
      le[e.posx][e.posy] = e;
      em1 = e;
      //If we are done set that flag as e=null.
      if (e != data.T) {e = events[++i];} else { e = null; }
    }

    //Add the rest, which was lying between events. Note that the lu thing has been advanced
    //be careful to add in right section.
    //System.out.println("Extra Adding " + betaHazAll(u,le) + " for time interval " + e.t + " to "  + em1.t);
    surv += betaHazAll(u,le) * (tstop-em1.t);

    return(-surv);
  }


  /** Abstract method to calculate the loglikelihood of the state.
   *  Parameterisation as I~\Gamma(gammaI,deltaI). The equation is taken from the article.
   *  Now extended to a spatial setting and optimized slightly for speed.
   *  The global variable @see useOrigins is used to check whether we use them in the calculations.
   *
   *  Note: It is required that genXYFuncs has been called.
   *  Internal non-documented options:
   *    useOrigins : Ignore the origin trick.
   */
  public double loglikGammaDelta() {
    /////////////////////////////////////////////////////////////////
    // Initialize data structure keeping track of the last event (le) in
    // each section.
    ////////////////////////////////////////////////////////////////
    Event le[][] = this.lastEventInit(true); //new array coz betaSurv uses it as well!!!

    //If there is a symantic error genXYZ set all X,Y,Z to NaN and
    //we should abort imediately and return an NaN.
    if (Double.isNaN(le[1][1].X)) { return(Double.NaN); }

    //////////////////////////////
    // exposedEvents
    //////////////////////////////
    double exposedLik = 0;
    //Start at event one coz e0 is number zero
    for (int i=1; i<noOfEvents; i++) {
      Event e = events[i];
      //Only interested in exposure events.
      if (e.type == 'E') {
        //System.out.println("Looking at the event: " + e);
        //System.out.println("Part1 = " + Math.log(betaHazAll(e.unit,le)));
        //System.out.println("Part 2 = " + betaSurv(e.unit,e.t));
        if (useOrigins) {
          exposedLik += Math.log(betaHaz(e.unit,e.origin,le));
        } else {
          exposedLik += Math.log(betaHazAll(e.unit,le));
        }

        //Show rotten events. But might be necessary if we propose them.
        if (LadyBug.DEBUG && ((exposedLik == Double.POSITIVE_INFINITY) || (Double.isNaN(exposedLik) || (exposedLik == Double.NEGATIVE_INFINITY)))) {
          System.err.println("Error in Event: :");
          System.err.println(e);
          System.err.println(this.toStringEvents());
          System.exit(-1);
        }

        //Add -int_{0}^{e_j} \lambda_all()
        exposedLik += betaSurvLn(e.unit,e.t);

        //System.out.println("exposedLik=" + exposedLik);
      }
      //Combinatorics
      //System.out.println("At t="+e.t+" multiplying by (+ on log)" + e.prevEventInSameUnit.X + " susceptibles");
      //exposedLik += Math.log(e.prevEventInSameUnit.X);

      //Advance events in section, where e happened. Not possible to advance from last T event.
      if (e.type != 'T') {
        //There appears to be a problem assigning using nextEventInSameUnit..
        //it's not always the same as e??!
        //le[e.posx][e.posy] = le[e.posx][e.posy].nextEventInSameUnit;
        le[e.posx][e.posy] = e;

      }
    } //end for

    ///////////////////////////////////////////////////////////////////
    // All individual events, i.e. recoveryEvents and possibly others
    ///////////////////////////////////////////////////////////////////
    double individualLik = 0;
    //double sumI=0;
    //double sumlogI=0;

    for (int i=0; i< data.noOfIndividuals; i++) {
      //Likelihood of all length of incubation periods. This contributes only
      //if incubation has a distribution
      if (Options.incuTime != 'C') {
        //Current hack to see what happens if the inoculated one is treated seperate.
        //if (i != data.inoculatedIndex) {
        individualLik += prob.loggammaPDF(I[i].t - E[i].t,gammaE,deltaE);
        //}
      }

      //Likelihood of all length of infection events.
      individualLik += prob.loggammaPDF(R[i].t - I[i].t,gammaI,deltaI);

      //Hook function for other events in addition to the recovery event.
      individualLik += loglikAddIndividualHook(i);

      if (LadyBug.DEBUG && ((individualLik == Double.POSITIVE_INFINITY) || (Double.isNaN(individualLik) || (individualLik == Double.NEGATIVE_INFINITY)))) {
        System.err.println("Error in Individual: Number " + i + " sucks.");
        System.err.println("E,I,R,D: =\n" + E[i] + "\n" + I[i] + "\n" + R[i]);
        try {
          System.err.println(((SEIRwDState)this).D[i]);
        } catch (Exception e) {
        }
        System.out.println("IndividualLik = " + individualLik);
        System.out.println("Exp  = " + prob.loggammaPDF(I[i].t - E[i].t,gammaE,deltaE));
        System.out.println("Inf  = " + prob.loggammaPDF(R[i].t - I[i].t,gammaI,deltaI));
        System.out.println("Extra= " + loglikAddIndividualHook(i));

        System.err.println(this.toStringEvents());
        System.exit(-1);
      }

    }
    //System.out.println("sumI=" + sumI + "\tsumlogI=" + sumlogI);

    //////////////////////////////////////////////////////////////////
    // The non-infected - REMEMBER: There should be a FUCKING MINUS.
    // That blasted&spoofed wrong signed costed 4 days in the UK
    // ladyBUG is definitely a suitable title!!!
    //////////////////////////////////////////////////////////////////
    double noinfLik = 0;
    for (int x=1; x<=data.noOfUnitsX; x++) {
      for (int y=1; y<=data.noOfUnitsY; y++) {
        noinfLik += data.units[x][y].STmembers * betaSurvLn(data.units[x][y],data.T.t);
      }
    }

    //Return the result
    //System.out.println("exposedLik  = " + exposedLik);
    //System.out.println("indivudalLik  = " + individualLik);
    //System.out.println("noinfLik  = " + noinfLik);
    //if (Double.isInfinite(exposedLik + recovLik + noinfLik)) {
    //  System.out.println("Probs");
    //}
    return(exposedLik + individualLik + noinfLik);
  }


  /** Just a dummy director to the currently used loglik function depending
   *  on the currently used parametrisation.
   */
  public double loglik() {
    double res = 0;
    //We'll use the mean,variance parametrisation
    if (MeanVarianceParameters) {
      //System.out.println("loglikMomement = " + loglikMoment());
      //System.out.println("loglikMomement2= " + loglikMoment2());
      //res = loglikMoment();
      //NOT IMPLEMENTED!!!
    } else {
      res = loglikGammaDelta();
    }
    return(res);
  }

  /** Get origins right once we know E(t), I(t), R(t), i.e can't pick an origin
   * where I(t) = 0.
   */
  public void sampleOrigins() {
    System.out.println("Sampling origins---------------->");

    //Init the last event structure
    Event le[][] = this.lastEventInit(false);
    Unit origin = null;

    //Loop over all events
    for (int i=0; i<noOfEvents; i++) {
      Event e = events[i];
      //Only exposures have an origin.
      if (e.type == 'E') {
        boolean validOrigin = false;
        //Sample origin among N_4^* until we find a valid one.
        while (!validOrigin) {
          origin =  e.unit.randomN4Star();
          validOrigin = le[origin.posx][origin.posy].Y > 0;
        }
        e.origin = origin;
      }
      //Advance and keep track of the last event we had a look at
      le[e.posx][e.posy] = e;
    }

    System.out.println("------- Result of sampleOrigin -------");
    for (int i=0; i<noOfEvents; i++) {
      Event e = events[i];
      //Only exposures have an origin.
      if (e.type == 'E') {
        System.out.println(e);
      }
    }
    System.out.println("------------------------");
  }

  /**
   * Hook function to initialize any additional parameters introduced in subclasses.
   *
   */
  protected void sampleStateParamsHook() {
    //Do nothing.
  }
  /**
   * Hook function to new any additional events in subclasses.
   * @param i Number of the individual we are newing events for.
   */
  protected void sampleStateEventsHook(int i, boolean sampleCensored) {
    //Do nothing
  }

  /** Make a valid instance of the parameters to start the MCMC chain with.
   *  If a proper prior is specified for the model parameters we begin in
   *  the mean of the prior. For I/R a state is samples for all parametric
   *  instances. Semantic is that if there is at least one missing R it's a transmission exp.
   *
   *   Note: Currently no regime or individual spatial param setup implemented. */
  public EpiState sampleState(boolean sampleCensored) {
    //Fetch initial values from startVal array.
    double theta0[] = new double[this.noOfModelParams+1];
    for (int i=1; i<=this.noOfModelParams; i++) {
      double val = data.startVal[i];
      if (!Double.isNaN(val)) {
        theta0[i] = val;
      } else {
        System.err.println("Currently not implemented.");
        System.exit(-1);
      }
    }

    //Translate array to parameters
    array2Params(theta0,false);

    //Debug
    theta0 = params2Array();
    System.out.println("Starting values: ");
    for (int i=1; i<=this.noOfModelParams; i++) {
      System.out.println(modelParamNames[i] + "\t=" + theta0[i]);
    }

    //Initialize any further parameters using a hook function.
    sampleStateParamsHook();


    //////////////////////////////////////////////////
    // I-values (it's really the l values)
    // Exposed and Origin events follow automatically
    /////////////////////////////////////////////////
    E = new Event[data.noOfIndividuals];
    I = new Event[data.noOfIndividuals];
    R = new Event[data.noOfIndividuals];

    boolean validConfiguration = false;

    ///////////////////////////////////////////////////////////////
    //This is the new way of creating a state. Here, the situation
    //of existing recoveries, but missing infection times is handled.
    ///////////////////////////////////////////////////////////////
    while (!validConfiguration) {
      //Loop over all events - infections as well as recoveries.
      for (int i=0; i<data.noOfIndividuals; i++) {
        //Copy-spam E,I,R events from the data file.
        E[i] = data.E[i];
        R[i] = data.R[i];
        I[i] = data.I[i];

        //If any of them is parametric, borrow starting value from Backup. This code currently
        //lacks functionality to sample a useful starting state by itself. Currently, we require
        //start value to be provided in data file.

        //It would be easier to treat E,I,R as a second dimension in the array.
        //Then we could make nicer code.

        //Initializing parametric E's
        if (E[i].isParametric) {
          if (!Double.isNaN(E[i].startT)) {
            E[i].t = E[i].startT;
          }
          else {
            System.err.println("Error: No starting value for E[" + i+"] found.");
            System.exit( -1);
          }
        }
      
      
        //If I is parametric we query the data file for a starting value. If there is no we are in trouble.
        if (I[i].isParametric) {
          if (!Double.isNaN(I[i].startT)) {
            I[i].t = I[i].startT;
          } else {
            System.err.println("Error: No starting value for I[" + i + "] found.");
            System.exit(-1);
          }
        }

        //In case incubation time is constant and not asis (i.e. incu=infinity) all incubation
         //times are set according to the I time.
         if ((Options.incuTime == 'C') && (!Double.isInfinite(data.incu))) {
           E[i].t = I[i].t - data.incu;
         } 
         
        //If R is parametric we query the data file for a starting value. (there won't be)
        //and if this does not exist we steal a value from the non parametric one.
        if (R[i].isParametric) {
          int j;
          do {
            j = prob.getPRNG().nextInt() % (data.noOfIndividuals);
          } while (data.R[j].isParametric); //warning. Not all R[j] might have been copied from data.
          R[i].t = I[i].t + (data.R[j].t-data.I[j].t);
        }

        //Hook function to create any additional events
        sampleStateEventsHook(i,sampleCensored);
      }

      //Debug stuff after sample state.
      for (int i=0; i<data.noOfIndividuals; i++) {
        System.out.println(i + " : " + E[i].t);
      }
      
      //Now we samples a state no matter if it is a transmission experiment
      //or not, now we need to create all the cool functions,
      //i.e. find S(t), E(t), I(t), and R(t)
      genXYFuncs();

      System.out.println("------\nAfter genXYZ in sampleState() function....");
      System.out.println(toStringEvents());

      //Set the origins to something more complicated than the initial section itself.
      sampleOrigins();
      /*System.out.println("First infected unit:\n"  + data.firstInfectedUnit);*/


      //Check if we samples a valid configuration.
      validConfiguration = this.isIValid();
      //validConfiguration=true;
    }

    paramList = new LinkedList();
    System.out.println("Event params = " + (noOfEParam + noOfIParam + noOfRParam));
    System.out.println("Number of Event params = " + paramList.size());

    //Setup list of all parametric events.
    for (int i=0; i<noOfIParam; i++) {
      paramList.add(I[IParamIndex[i]]);
    }
    for (int i=0; i<noOfRParam; i++) {
      paramList.add(R[RParamIndex[i]]);
    }
    //Add E to the list of all parametric events. These have acceptance rate etc.
    for (int i=0; i<noOfEParam; i++) {
      paramList.add(E[EParamIndex[i]]);
    }
    System.out.println("Event params = " + (noOfEParam + noOfIParam + noOfRParam));
    System.out.println("Number of Event params = " + paramList.size());

    System.out.println("We got a valid configuration!");
    return(this);
  }


  /**
   * Random Walk Metropolis-Hastings updating of a parameter with gamma prior
   * distribution. This is general method not necessarily efficient as the entire
   * likelihood is calculated.
   *
   * @param theta Vector of all parameters.
   * @param paramIndex of the parameter to update
   * @param sigma How large a standard deviation to use in the proposal distribution
   *
   * Note: The value is updated automatically by calling the array2Param function
   * Therefore no value is returned
   */

  public void updateMetroHastings(double theta[], int paramIndex, double sigma) {
    //Current value
    double paramVal = theta[paramIndex];

    //Find likelihood of current value * prior
    double logfold = loglik() + (data.nu[paramIndex]-1) * Math.log(theta[paramIndex]) - data.lambda[paramIndex]*theta[paramIndex]; ;

    //Make a deltaI step using Random walk with step size specified in options.
    theta[paramIndex] = Math.abs(theta[paramIndex] + prob.normalRand()*sigma);

    //System.out.println("Proposing deltaI=" + deltaI + " instead of " + savedeltaI);
    //Check new loglik - f(R,I-beta_{xyr}+deltaI_{xyr}')
    array2Params(theta,false);
    double logfnew = loglik() +   (data.nu[paramIndex]-1) * Math.log(theta[paramIndex]) - data.lambda[paramIndex]*theta[paramIndex]; ;

    //Accept the new value y/n? No need for genXYFunc coz we have not moved any of these values.
    if (!acceptProposal(logfnew,logfold)) {
      //Go back
      theta[paramIndex] = paramVal;
      array2Params(theta,false);
    } else { //accept
      accept[paramIndex]++;
    }
  } //end of metropolis hasting

  /**
   * Gibbs updating of beta - we know the full conditional (see paper)
   * This version handles the multiple unit setup.
   */
  public void updateBeta() {
    if (useOrigins) {
      /////////////////////////////////////
      // a = |E_{loc}| + \nu_\beta
      //////////////////////////////////////
      LinkedList Eloc = Eloc();
      double a = Eloc.size() + data.nu[beta_];

      ///////////////////////////////////////
      // b = \lambda_beta + all the local ones + all survivors.
      ///////////////////////////////////////
      double b = data.lambda[beta_];

      // \sum_{j\in A_T\backslash\{1\}} \int_{i_0}{e_j} \sum_{u\in N(u_j)} I_u(t) dt
      for (int i=0; i<data.noOfIndividuals; i++) {
        if (i!=data.inoculatedIndex) {
          Event e = E[i];
          b += Yintegrate(e.unit,e.t);
        }
      }
      // \sum_{e\in E_{loc}} \int_{i_0}{e} Y_{e.u}(t) dt
      /*Iterator iter = Eloc.iterator();
      while(iter.hasNext()) {
        Event e = (Event)iter.next();
        b += Yintegrate(e.unit,e.t);
      }*/

      // sum_{j\in S_T} \int_{i_0}^{T} Y_{u_j}(t) dt
      for (int x=1; x<=data.noOfUnitsX; x++) {
        for (int y=1; y<=data.noOfUnitsY; y++) {
          b += data.units[x][y].STmembers * Yintegrate(data.units[x][y],data.T.t);
        }
      }
      //System.out.println("a=" + a + "\tb=" + b);
      beta = prob.gammaRand(a,b);
      //Always accept
      accept[beta_]++;
    } else { //Metropolis update
      theta = params2Array();
      updateMetroHastings(theta,beta_,Options.mcmc_betaRWsigma);
    }
  }

  /**
   * Gibbs updating of betaN - we know the full conditional (see paper)
   * This is a new version correcting an mistake of the conditioning on o_j.
   */
  public void updateBetaN() {
    if (useOrigins) {
      /////////////////////////////////////
      // a = |E_{nbr}| + \nu_\betaN
      //////////////////////////////////////
      LinkedList Enbr = Enbr();
      double a = Enbr.size() + data.nu[betaN_];

      ///////////////////////////////////////
      // b = \lambda[betaN_] + all the local ones + all survivors.
      ///////////////////////////////////////
      double b = data.lambda[betaN_];

      // \sum_{j\in A_T\backslash\{1\}} \int_{i_0}{e_j} \sum_{u\in N(u_j)} I_u(t) dt
      for (int i=0; i<data.noOfIndividuals; i++) {
        if (i!=data.inoculatedIndex) {
          Event e = E[i];
          LinkedList N4 = e.unit.N4();
          Iterator uiter = N4.iterator();
          while (uiter.hasNext()) {
            b += Yintegrate((Unit)uiter.next(),e.t);
          }
        }
      }

      // sum_{j\in S_T} \int_{i_0}^{T} \sum{u'\in N_4(u_j)} Y_{u}(t) dt
      for (int x=1; x<=data.noOfUnitsX; x++) {
        for (int y=1; y<=data.noOfUnitsY; y++) {
          LinkedList N4 = data.units[x][y].N4();
          Iterator iter = N4.iterator();
          while (iter.hasNext()) {
            b += data.units[x][y].STmembers * Yintegrate((Unit)iter.next(),data.T.t) ;
          }
        }
      }
      //System.out.println("betaN:\na=" + a + "\tb=" + b);
      betaN = prob.gammaRand(a,b);
      //Always accept
      accept[betaN_]++;
    } else {
      theta = params2Array();
      updateMetroHastings(theta,betaN_,Options.mcmc_betaNRWsigma);
    }
  }

  /**
   * Gibbs updating of deltaI - we know the full conditional (see paper)
   */
  public void updateDeltaE() {
    //If this period is constant then we forget all about it.
    if (Options.incuTime == 'C') {deltaE = 0; return;}
    //In case we wan to use the general Metropolis-Hastings function
    //if (Options.slowButSafe) {
    //  double theta[] = this.params2Array();
    //  updateMetroHastings(theta,deltaE_, Options.mcmc_deltaERWsigma);
    //  return;
    //}

    //inoculated is part otherwise use (data.noOfIndividuals-1)
    double a = (data.noOfIndividuals) * gammaE + data.nu[deltaE_];

    // b = \sum_{i=1}^N (r_j - i_j) + \lambda_deltaI
    double b = data.lambda[deltaE_];
    for (int i=0; i<data.noOfIndividuals; i++) {
      //Inoculated not part
      //if (i!=data.inoculatedIndex) {
      b += (I[i].t - E[i].t);
      //}
    }
    deltaE = prob.gammaRand(a,b);
    //GibbsUpdating always accepts
    accept[deltaE_]++;
  }


  /**
   * Gibbs updating of deltaI - we know the full conditional (see paper)
   */
  public void updateDeltaI() {
    //Use general Metropolis Hastings function (safe)
    /*if (Options.slowButSafe) {
      theta = params2Array();
      updateMetroHastings(theta,deltaI_,Options.mcmc_deltaIRWsigma);
      return;
    }*/

    // a = N*gammaI + nu[deltaI_]
    double a = data.noOfIndividuals * gammaI + data.nu[deltaI_];

    // b = \sum_{i=1}^N (r_j - i_j) + \lambda[deltaI_]
    double b = data.lambda[deltaI_];
    for (int i=0; i<data.noOfIndividuals; i++) {
      b += (R[i].t - I[i].t);
    }
    deltaI = prob.gammaRand(a,b);
    //GibbsUpdating always accepts
    accept[deltaI_]++;
  }


  /**
   * Update gammaE using Metropolis Hastings. This is a new improved version only using
   * the relevant parts of loglik and prior. I.e. \pi(\gamma_E|\delta_E,D)
   * and equation about full condition in article. As noted we could also
   * try to do this using adaptive rejection sampling (ARS).
   * Note though mean/var representation does not work now, but it sucks anyway.
   *
   */
  public void updateGammaE() {
    //If we specified using an exponential distribution we do not need to do
    //anything.
    if (Options.incuTime == 'E') {gammaE = 1; return;}
    if (Options.incuTime == 'C') {gammaE = 0; return;}
    if (Options.slowButSafe) {
      double theta[] = this.params2Array();
      updateMetroHastings(theta,gammaE_, Options.mcmc_gammaERWsigma);
      return;
    }

    //Calculate log Y = \sum_{j\in A_T} log(i_j-e_j)
    //Note: Consider, whether the inoculated invidual should be part of the model...
    double logY = 0;
    for (int i=0; i< data.noOfIndividuals; i++) {
      //if (i!=data.inoculatedIndex) {
      logY += Math.log(I[i].t - E[i].t);
      //}
    }

    //Find value of proportional likelihood part evaluated at old value use (data.noOfIndividuals-1) if first is not part
    double loglikold  = data.noOfIndividuals * gammaE * Math.log(deltaE) + (gammaE-1)*logY - data.noOfIndividuals * prob.lngamma(gammaE);
    // Posterior evaluated at old value: lik \times prior
    double logfold = loglikold + (data.nu[gammaE_]-1) * Math.log(gammaE) - data.lambda[gammaE_]*gammaE;

    //Save value
    double savegammaE = gammaE;
    //Make a gammaE step using Random walk - use abs(val) - property of Gaussian distribution
    gammaE = Math.abs(gammaE + prob.normalRand()*Options.mcmc_gammaERWsigma);

    //Check new loglik - f(R,I-gammaE_{xyr}+gammaE_{xyr}') use (data.noOfIndividuals-1) if first is not part
    double logliknew  = data.noOfIndividuals * gammaE * Math.log(deltaE) + (gammaE-1)*logY - data.noOfIndividuals * prob.lngamma(gammaE);
    // Posterior evaluated at new value: lik \times prior
    double logfnew = logliknew + (data.nu[gammaE_]-1) * Math.log(gammaE) - data.lambda[gammaE_]*gammaE;

    //Accept the new value y/n? No need for genXYFunc coz we have not moved any of these values.
    if (!acceptProposal(logfnew,logfold)) {
      gammaE = savegammaE;
    } else { //accept
      accept[gammaE_]++;
    }
  }

  /** Update gammaI using Metropolis Hastings. This new version includes only relevant
   * parts of the likelihood when calculating the acceptance prob. Quite a speed gain!
   * Note though mean/var representation does not work now, but it sucks anyway.
   */
  public void updateGammaI() {
    //If we specified using an exponential distribution we do not need to do anything.
    if (Options.expInfTime) {gammaI = 1; return;}
    //Use safe general purpose Metropolis Hastings update
    if (Options.slowButSafe) {
      double theta[] = this.params2Array();
      updateMetroHastings(theta,gammaI_, Options.mcmc_gammaIRWsigma);
      return;
    }

    //Calculate log Y = \sum_{j\in A_T} log(r_j-i_j)
    double logY = 0;
    for (int i=0; i< data.noOfIndividuals; i++) {
      logY += Math.log(R[i].t - I[i].t);
    }

    //Find value of proportional likelihood part evaluated at old value
    //double loglikold  = loglik();
    double loglikold  = data.noOfIndividuals * gammaI * Math.log(deltaI) + (gammaI-1)*logY - data.noOfIndividuals * prob.lngamma(gammaI);
    // Posterior evaluated at old value: lik \times prior
    double logfold = loglikold + (data.nu[gammaI_]-1) * Math.log(gammaI) - data.lambda[gammaI_]*gammaI;

    //Make a gammaI step using Random walk
    double sigma = Options.mcmc_gammaIRWsigma; //.5; //This parameter is open for tuning!
    double savegammaI = gammaI;
    // Make proposal value gammaI'_{r} - use abs(val) - property of Gaussian distribution
    gammaI = Math.abs(gammaI + prob.normalRand()*sigma);
    //deltaI = gammaI*gammaI; //they are locked together.
    //gammaI = gammaI + prob.normalRand()*sigma;

    //Check new loglik - f(R,I-gammaI_{xyr}+gammaI_{xyr}')
    //Find value of proportional likelihood part evaluated at new value - Note: Only relevant part instead of calling loglik()
    //double logliknew  = loglik();
    double logliknew  = data.noOfIndividuals * gammaI * Math.log(deltaI) + (gammaI-1)*logY - data.noOfIndividuals * prob.lngamma(gammaI);
    // Posterior evaluated at new value: lik \times prior
    double logfnew = logliknew + (data.nu[gammaI_]-1) * Math.log(gammaI) - data.lambda[gammaI_]*gammaI;

    //Accept the new value y/n? No need for genXYFunc coz we have not moved any of these values.
    if (!acceptProposal(logfnew,logfold)) {
      gammaI = savegammaI;
      //deltaI = gammaI*gammaI;
    } else { //accept
      accept[gammaI_]++;
    }
  }


  /** Update the o vector using Gibbs Sampling. This implementation is not very speedy and uses established \lambda
   *  functions instead of doing the necessary integration itself. */
  public void updateO() {
    //System.out.println(toStringEvents());

    double p[] = new double[5];
    Unit   u[] = new Unit[5];

    //Initialize structure. No need to new the structure, coz we do not call
    //in any of the sub routines.
    Event le[][] = this.lastEventInit(false);

    //Now increase and act on exposed events.
    for (int i=1; i<noOfEvents; i++) {
      Event e = events[i];
      //Only interested in exposed events.
      if (e.type == 'E') {
        //Sum of probs
        double sum = 0;
        //Loop over all neighbours'
        LinkedList N4Star = e.unit.N4Star();
        Iterator iter = N4Star.iterator();
        for (int no = 0; no < N4Star.size(); no++) {
          u[no] = (Unit)iter.next();
          p[no] = betaHaz(e.unit,u[no],le);
          sum += p[no];
        }
        //Sample according to the above discrete distribution
        double r = prob.uniformRand()*sum;
        int cell = -1;
        while (r<=sum) {
          cell++;
          sum -= p[cell];
        }
        //Assign the new origin value
        e.origin = u[cell];
      } //end of exposed if
      //Advance events in section, where e happened. Not possible to advance from last T event.
      if (e.type != 'T') {
        //Appear to be a problem with nextEventInSameUnit...when is it not equal to??
        //le[e.posx][e.posy] = le[e.posx][e.posy].nextEventInSameUnit;
        le[e.posx][e.posy] = e;
      }
    }//end of for all events
  } //end of method


  /** Update E using full conditionals. Currently, Metropolis Hastings is used, but
   * at this point we would very much prefer to use some adaptive rejection scheme.
   * Still, the formula is quicker to update so we update all at the same time.
   * This version does not work for mean/var parametrization, just the natural.
   * Note: This does not take Diagnostic events into account. Specialized version necessary!
   * */
  public void updateE() {
    //Loop over all E in EParamIndex to updateIn case E is fixed nothing needs to be done
    //Make sure e1 is not parametric!
    int batchAccept = 0;

    //Loop over all E events who need to be updated
    for (int m=0; m<this.noOfEParam; m++) {
      //Find the index we want.
      int j = this.EParamIndex[m];

      //For each we (unfortunately need to loop in the chain to the event
      //so the lastEvent structure is correct. Snapshots would be great here!

      /////////////////////////////////////////////////////////////////
      // Initialize data structure keeping track of the last event (le) in
      // each section.
      ////////////////////////////////////////////////////////////////
      Event le[][] = this.lastEventInit(true); //new array coz betaSurv uses it as well!!!

      //Loop correctly updating le until we find the event of interest.
      int i=0;
      //System.out.println("Looking at E["+j+"]="  + E[j].t);
      while (events[i] != E[j]) {
        Event e = events[i];
        le[e.posx][e.posy] = e;
        i++;
      }

      //Save its stats - i.e. time and current likelihood*prior
      double saveE = E[j].t;

      //Calculate w(X) = pi(x)/q(X) according to p.9 in the Gilks book. Note: Proposal q(X)
      //is of independence sampler type, hence we use eq. (1.8). Not symmetric!!
      double loglikOld = Math.log(betaHaz(E[j].unit,E[j].origin,le)) + betaSurvLn(E[j].unit,E[j].t);
      double logWOld = loglikOld + (gammaE-1)*Math.log(I[j].t-E[j].t) - deltaE*(I[j].t-E[j].t); // - prob.loggammaPDF(I[j].t-E[j].t,gammaE,deltaE);


      //Propose new value, use q(), which is a good proposal density.
      //Coz we will reject all exposure times before e1=0 we loop
      //until we get a valid one. Is this valid?
      double foo;
      do {
        //foo = I[j].t - prob.gammaRand(gammaE,deltaE);
        foo  = E[j].t + prob.normalRand()*Options.mcmc_ERWsigma;
      } while ((foo<=0) || (foo>=I[j].t));
      E[j].t = foo;


      //To make sure time order is still sorted, we better recompute all events.
      genXYFuncs();
      //Move up the list again to get right neighbour values.
      le = this.lastEventInit(true); //new array coz betaSurv uses it as well!!!
      //Loop correctly updating le until we find the event of interest.
      i=0;
      while (events[i] != E[j]) {
        Event e = events[i];
        le[e.posx][e.posy] = e;
        i++;
      }
      //System.out.println("|Enbr|=" + this.Enbr().size());
      ////////////////////////////////////////////////////////////////////////

      //Calculate w(Y) = pi(Y)/q(Y).
      double loglikNew = Math.log(betaHaz(E[j].unit,E[j].origin,le)) + betaSurvLn(E[j].unit,E[j].t);
      double logWNew = loglikNew + (gammaE-1)*Math.log(I[j].t-E[j].t) - deltaE*(I[j].t-E[j].t); // - prob.loggammaPDF(I[j].t-E[j].t,gammaE,deltaE);

      //Check whether proposal is accepted and do appropriate action.
      if (!acceptProposal(logWNew,logWOld)) {
        //Restore the old value and restore all events (might be superflous)
        E[j].t = saveE;
        genXYFuncs();
      } else {
        batchAccept++;
        E[j].accepted++;
      }
    }
    //System.out.println("batchAccept = " +  batchAccept);
    //Add to acceptance rates.
    Eaccept += batchAccept/(double)noOfEParam;
  }

  /** Update I using Metropolis Hastings. */
  public void updateI() {
    //Do we have something to do?
    if (noOfIParam==0) {return;}

    //Count how many proposals of this batch are successfully accepted.
    int batchAccept = 0;

    //Loop over all parameters to update instead of IParamIndex[prob.getPRNG().nextInt() % (noOfIParam)];
    for (int i=0; i<noOfIParam; i++) {
      int j=IParamIndex[i];

      //Save its stats - , i.e. time and current likelihood*prior as well as the corresponding exposure event?
      double currentIt = I[j].t;
      double currentEt = E[j].t;

      //Calculate w(X) = pi(x)/q(X) according to p.9 in the Gilks book. Note: Proposal q(X)
      //is of independence sampler type, hence we use eq. (1.8). Not symmetric!!
      double logWOld = loglikGammaDelta() - prob.loggammaPDF(R[j].t-I[j].t,gammaI,deltaI);

      //Propose new value. No tuning parameter here!
      I[j].t = R[j].t - prob.gammaRand(gammaI,deltaI);
      //In case incubation time is fixed this has to be taken into account!
      if (Options.incuTime == 'C') {E[j].t = I[j].t-data.incu;}
      //In case a Random Walk update is prefered -- currently not the case.
      //I[j].t = Math.abs(I[j].t + prob.normalRand()*Options.mcmc_IRWsigma);

      //Calculate probability of the proposal. Need to sort all events again using genXYZ
      genXYFuncs();
      double logWNew = loglikGammaDelta() - prob.loggammaPDF(R[j].t-I[j].t,gammaI,deltaI);

      //Check whether proposal is accepted and do appropriate action.
      if (!acceptProposal(logWNew,logWOld)) {
        //Restore the old value
        I[j].t = currentIt;
        E[j].t = currentEt;
        genXYFuncs();
      } else {
        batchAccept++;
        //New individual accept monitoring
        I[j].accepted++;
      }
    }
    Iaccept +=  (double)batchAccept/(double)noOfIParam;
  }

  /** Update R using Metropolis Hastings. */
  public void updateR() {
    //Do we have something to do?
    if (noOfRParam==0) {return;}

    int batchAccept = 0;
    //Loop over all parametric recovery events, instead of RParamIndex[prob.getPRNG().nextInt() % (noOfRParam)];
    for (int i=0; i<noOfRParam; i++) {
      int j = RParamIndex[i];
      //Save its stats - , i.e. time and current likelihood*prior as well as the corresponding exposure event?
      double currentRt = R[j].t;

      //Calculate w(X) = pi(x)/q(X) according to p.9 in the Gilks book.
      //Note: Proposal q(X) is symmetric (gaussian) so we need no correction
      double logWOld = loglikGammaDelta();

      //Propose new value from a Gaussian. We know R[i].t>0, but we also know R[i].t<T
      double foo;
      do {
        foo = Math.abs(R[j].t + prob.normalRand()*Options.mcmc_RRWsigma);
      } while (foo>data.T.t);
      R[j].t = foo;
      //R[j].t = Math.abs(R[j].t + prob.normalRand()*Options.mcmc_RRWsigma);

      //Calculate probability of the proposal. Need to sort all events again using genXYZ
      genXYFuncs();
      double logWNew = loglikGammaDelta();

      //System.out.println("New: I["+i+"]=" +  LadyBug.form.format(I[i].t) + "\tloglik=" +  LadyBug.form.format(loglik()) + "\tloggammaPDF=" +  LadyBug.form.format(prob.loggammaPDF(data.R[i].t-I[i].t,1,gammaI)));
      //System.out.println(logWNew - logWOld);

      boolean accept = acceptProposal(logWNew,logWOld);

      //Check whether proposal is accepted and do appropriate action.
      if (!accept) {
        //Restore the old value
        R[j].t = currentRt;
        genXYFuncs();
      } else {
        batchAccept++;
        R[j].accepted++;
      }
    }
    Raccept += batchAccept/(double)this.noOfRParam;
  }

  /**
   * Hook method to update additional parameters using MCMC. To be specialized by subclass.
   * @param justTheta no data just parameters are updated.
   */
  public void updateHook(boolean justTheta) {
    //Do nothing.
  }

  /** Generate method for MCMC chain
   *  Modifcation happen on the instance itself to keep things simple.
   *
   *   @return this */
  public EpiState generateNext(boolean justTheta) {
    //Trial:
    //useOrigins = false;

    long total = Options.mcmc_samples*Options.mcmc_thin + Options.mcmc_burnin;
    if ((runs % Math.floor(total/100))  == 0) {
      System.out.print(runs*100/total + "%");
      //Every 10%
      if ((runs % Math.floor((total*10)/100))  == 0) {
        //Show status
        System.out.println(toStringEvents());
        System.out.println(toStringLong());
      }
    }
    runs++;

    ///////////////////////////////////////////
    // Update all parameters (many!)
    //Transmission part
    updateBeta();

    //Spatial updates if necessary.
    if (data.noOfUnits>1) {
      updateBetaN();
      updateO();
      //System.out.println("|Eloc| = " + Eloc().size() + "\t" + "|Enbr| = " + Enbr().size());
    }

    //Length of incubation period
    updateGammaE();
    updateDeltaE();

    //Length of infectious period
    updateGammaI();
    updateDeltaI();

    if (!justTheta) {
      //We assume only E's can be updated.
      updateE();
      //Missing I's?
      updateI();
      updateR();
    }
    
    //Update any additional parameters.
    updateHook(justTheta);

    //Security check to ensure that loglik isn't becoming bogus. While we are
    //at it we can just as well set the loglik variable, so we do not need to
    //recompute the value while writing out status information.
    loglik = loglik();
    if (Double.isInfinite(loglik)) {
      //Turn debug mode on so we are showing which event/individual makes
      //things become bogus.
      LadyBug.DEBUG = true;
      loglik();
    }

    return(this);
  }

  /**
   * Hook function to move additional parameters in numerical optimizer.
   *
   * @param theta The theta array
   */
  public void f_to_minimizeHook(double[] theta) {
    //Do nothing.
  }

  /** Small stats function to compute mean and variance for inf or inc periods
   * of current configuraion
   *
   * @return 4Dim array containing E_e, E_i, V_e, V_i
   */
  public double[] periodsEV() {
    double EV[] = new double[4];

    //Calculate mean of inc and inf periods, respectively
    for (int i=0; i<data.noOfIndividuals; i++) {
      EV[0] += (I[i].t - E[i].t);
      EV[1] += (R[i].t - I[i].t);
    }
    EV[0] /= data.noOfIndividuals;
    EV[1] /= data.noOfIndividuals;

    //Calculate variances
    for (int i=0; i<data.noOfIndividuals; i++) {
      EV[2] += Math.pow((I[i].t - E[i].t) - EV[0],2);
      EV[3] += Math.pow((R[i].t - I[i].t) - EV[1],2);
    }
    EV[2] /= (data.noOfIndividuals-1);
    EV[3] /= (data.noOfIndividuals-1);

    //Return the result
    return(EV);
  }
  /** Helper functions to put all named parameters in an array. Remember Fortran indexing
   * Function currently does not work if incu period a constant.
   *
   * @return An @see noOfModelParams + 1 long array with params
   */
  public double[] params2Array() {
    double theta[] = new double[noOfModelParams+1];
    //Assign all model parameters.
    theta[beta_]   = beta;
    theta[betaN_]  = betaN;
    theta[gammaE_] = gammaE;
    theta[deltaE_] = deltaE;
    theta[gammaI_] = gammaI;
    theta[deltaI_] = deltaI;

    return(theta);
  }
  /** Helper function, given an array of size noOfModelParams assign
   * the values of this vector to the correct names.
   *
   * @param theta - A vector of length noOfModelParams containing the values to assign
   * @param expTransformParams - When using the minimizer it is a good idea to operate with
   *                             log(param) value instead to avoid negative values.
   */
  public void array2Params(double theta[], boolean expTransformParams) {
    //Infectious period
    deltaI = expTransformParams ? Math.exp(theta[deltaI_]) : theta[deltaI_];
    if (Options.expInfTime) {
      gammaI = 1;
    } else {
      gammaI = expTransformParams ? Math.exp(theta[gammaI_]) : theta[gammaI_];
    }

    //Transmission
    beta   = expTransformParams ? Math.exp(theta[beta_]) : theta[beta_];
    if (data.noOfUnits > 1) {
      betaN   = expTransformParams ? Math.exp(theta[betaN_]) : theta[betaN_];
    } else {
      betaN = 0;
    }

    //Incubation time if necessary.
    if (Options.incuTime != 'C') {
      if (Options.incuTime == 'E') {
        gammaE = 1;
      } else {
        gammaE = expTransformParams ? Math.exp(theta[gammaE_]) : theta[gammaE_];
      }
      deltaE = expTransformParams ? Math.exp(theta[deltaE_]) : theta[deltaE_];
    }
  }

  /** In case the object is given to a numerical minimizer, this is the function to minimize.
   *  We will only do maximization of the likelihood, i.e. parameters have to be
   *  transmoglyffed to account for different indexing of arrays etc.
   *  Thus, this will work as a wrapper to logflambdaI.
   *
   *  @param  theta The parameter vector(remember Fortran indexing of arrayx (starting at 1))
   */
  public double f_to_minimize(double theta[]) {
    //Implement constraint that gammaI>0 and beta>0. Works both if they are parameters
    //in a gamma distribution or mean/variance.

    //Assign theta vector to each parameter. Depending on global parameter expTransformParams
    //subclasses should overwrite this method.
    array2Params(theta,expTransformParams);

    //Specialty of the meanvariance transformation is that in a ML
    //context this is not an operation on random variables.
    //I.e. no need to correct for the transformation! And we hence transform
    //without performing any type of correction.

    double loglik = 0;
    if (this.MeanVarianceParameters) {
      System.err.println("Not implemented!!!!!!!!!!!!!!!!!");
      System.exit(-1);
    } else {
      //Just use diagnostic times
      if (justDiagnostic) {
        for (int i=0; i<data.noOfIndividuals; i++) {
          loglik += loglikAddIndividualHook(i);
        }
      } else {
        loglik = loglikGammaDelta();
      }
    }

    //Return result.
    return(-loglik);
  }

  /** Gradient function, in case one wants to do numeric optimization of the likelihood.
   *  Contains an EMPTY body.
   *
   *  @param x The parameter (remember Fortran indexing of arrayx (starting at 1))
   *  @param g The computed gradient.
   */
  public void gradient(double x[], double g[]) {}

  /** Hessian function, in case one wants to do numeric optimization of the likelihood.
   *  Contains an EMPTY body.
   *
   *  @param x The parameter (remember indexing of arrayx (starting at 1))
   *  @param g The computed gradient.
   */
  public void hessian(double x[], double h[][]) {};

  /** Hook function returning how many additional events are put into the
   * events array by subclasses etc. To be rewritten by subclasses.
   *
   * @return Number of additional events added.
   */
  protected int extraNoOfEventsHook() {
    return 0;
  }
  /** Hook function to add additional events to the events array.
   *  To be rewritten by subclasses.
   *
   * @param baseNoOfEvents Number of E,I,R events in SEIRState class.
   * @return Number of additional events added.
   */
  protected void addExtraEventsHook(int baseNoOfEvents) {
    // Do nothing.
  }
  /** Hook function for adding additional events to the loglikelihood. It is called
   * from within loglikGammaDelta. The recovery events are already added. To be specialized
   * in subclasses.
   *
   * @param individual The i'th individual to add events for
   * @return Likelihood of the additional events.
   */
  protected double loglikAddIndividualHook(int individual) {
    //Do nothing.
    return 0;
  }

  /**
        Generate X, Y, XY, and time by sorting appropriate events.
        Note: New Version using Event classes.
        Before, we sorted by zipping. Now we use general sort on all events. Slower?
        Probably - use zipper again, which required I to be sorted, but then we have to sort
        it twice??
        */
  public void genXYFuncs() {
    //We are creating.
    nogen = false;

    //Deduce correct values X (susceptibles) and Y (infected) over time
    //For this we have arrays to hold all the f(t), step functions.
    //Note: We add one extra event to cope for the 'T' event (last time point).
    //      Assumption: T>last event.
    //No of events: I,R events, and (E-1) (E[inoculatedIndex] is not added)
    int baseNoOfEvents = data.noOfIndividuals+2*data.noOfIndividuals-1;
    //Additional hook events (e.g. Diagnostic) and the T event
    noOfEvents = baseNoOfEvents + extraNoOfEventsHook() + 1;

    events  = new Event[noOfEvents];

    //Make a pool of infections (from state).
    System.arraycopy(this.I,0,events,0,data.noOfIndividuals); //no need for this.I to be sorted?!?
    System.arraycopy(this.R,0,events,data.noOfIndividuals,data.noOfIndividuals);

    ///////////////////////////////////////////////////////////////////////////////////
    //Add all the exposures - except the one belonging to inoculatedIndex coz this has no exposure
    ///////////////////////////////////////////////////////////////////////////////////
    int e0index = 0;
    for (int i=1; i<data.noOfIndividuals; i++) {
      if (E[i].t < E[e0index].t) { e0index = i; }
    }
    //E:Copy from 0 to e0index-1 (problem if e0index==0?)
    if (e0index != 0) {
      System.arraycopy(this.E,0,events,data.noOfIndividuals+data.noOfIndividuals,e0index);
    }
    //E:Copy from e0index+1 to data.numberOfInfections-1
    System.arraycopy(this.E,e0index+1,events,data.noOfIndividuals+data.noOfIndividuals + e0index,data.noOfIndividuals-1-e0index);

    ///////////////////////////////////////////////////////////////////////////
    // Space for adding addition events in subclasses using a hook function
    //////////////////////////////////////////////////////////////////////////
    addExtraEventsHook(baseNoOfEvents);

    //Now add the T event - it's the same every time!
    events[noOfEvents-1] = data.T;

    //System.out.println("Events before sorting: ");
    //for(int i=0; i<noOfEvents; i++) {
    //  System.out.println(events[i]);
    //}

    //Sort the pool using event as Comparator. Is this slower than using a manual zip
    //fit, which could exploit that some of the arrays (e.g. recovery times) are already sorted?
    java.util.Arrays.sort(events,new Event());

    ///////////////////////////////////////////////////////////////
    //We are now ready to deduce X(t) and Y(t) for each unit, etc.
    //////////////////////////////////////////////////////////////

    //For bookkeeping
    Event lastEventInUnit[][] = new Event[data.noOfUnitsX+1][data.noOfUnitsY+1];

    //Assumption: I_0 is the first event. This is not necessary the case if we (by accident)
    //moved I0 ahead so there is another exposure coming first. Hence we just operate
    //with events[0] as the first event, not necessary being an infection?
    for (int x=1; x<=data.noOfUnitsX; x++) {
      for (int y=1; y<=data.noOfUnitsY; y++) {
        Unit u = data.units[x][y];
        //The I_0 event of course goes into data.firstUnit, but
        //for the others we generate an artificial event loaded with (N,a) information.
        if (u == data.firstInfectedUnit) {
          lastEventInUnit[x][y] = events[0]; //superflous?
          firstEventInUnits[x][y] = events[0];
          u.firstEvent = events[0];
        } else {
          double startEvTime[] = new double[2];
          startEvTime[0] = Double.NaN;
          startEvTime[1] = events[0].t;
          Event startEv = new Event(u,startEvTime,'A');
          lastEventInUnit[x][y] = startEv;
          firstEventInUnits[x][y] = startEv;
          u.firstEvent = startEv;
        }
        //Load (N,a) information, if the first event is an infection. Otherwise
        //make sure we only have crap in here so we definitely will reject such
        //a proposal!! Easier to write crap than make all functions aware
        //that the first event might not be an infection.
        if (events[0].type == 'I') {
          lastEventInUnit[x][y].X = data.units[x][y].N;
          lastEventInUnit[x][y].Y = data.units[x][y].a;
          lastEventInUnit[x][y].Z = 0;
        } else {
          //System.out.println("Trying to move beyond I0");
          lastEventInUnit[x][y].X = Double.NaN;
          lastEventInUnit[x][y].Y = Double.NaN;
          lastEventInUnit[x][y].Z = Double.NaN;
        }
      }
    }
    ///////////////////////////////////////////////////////////////
    //Loop over all remaining events, i.e. everything but I_0.
    /////////////////////////////////////////////////////////////////
    for (int i=1; i<noOfEvents; i++) {
      Event e = events[i];

      switch (e.type) {
        //If it's a 'I' or 'R' event it goes within this unit (local thing)
        case 'I':
        case 'R':
        case 'E':
        case 'D':
          if (e.type == 'E') { //exposures
            e.X = lastEventInUnit[e.posx][e.posy].X-1;
            e.Y = lastEventInUnit[e.posx][e.posy].Y;
            e.Z = lastEventInUnit[e.posx][e.posy].Z+1;
          } else if (e.type == 'I') { //infections
            e.X = lastEventInUnit[e.posx][e.posy].X;
            e.Y = lastEventInUnit[e.posx][e.posy].Y+1;
            e.Z = lastEventInUnit[e.posx][e.posy].Z-1;
          } else if (e.type == 'R') { //recoveries
            e.X = lastEventInUnit[e.posx][e.posy].X;
            e.Y = lastEventInUnit[e.posx][e.posy].Y-1;
            e.Z = lastEventInUnit[e.posx][e.posy].Z;
          } else if (e.type == 'D') { //diagnosed
            e.X = lastEventInUnit[e.posx][e.posy].X;
            e.Y = lastEventInUnit[e.posx][e.posy].Y;
            e.Z = lastEventInUnit[e.posx][e.posy].Z;
          }
          //Previous event should point to current lastEventInUnit event
          e.prevEventInSameUnit = lastEventInUnit[e.posx][e.posy];
          //The previous event should have its next event point to e
          e.prevEventInSameUnit.nextEventInSameUnit = e;
          //We can now advance the lastEvent structure.
          lastEventInUnit[e.posx][e.posy] = e;
          break;
          //If it's the terminating event, we clone one for each unit and make it the units
          //last event. It might not be necessary to actually clone one - could
          //make pointers to the T, but the clone solution is easier
          //as all events can be treated equal then.
        case 'T':
          for (int x=1; x<=data.noOfUnitsX; x++) {
            for (int y=1; y<=data.noOfUnitsY; y++) {
              //Be aware that the event is also in the events array, which
              //should be updated as well!
              Event unitSEv;
              if (e.unit != data.units[x][y]) {
                double unitSEvTime[] = new double[2];
                unitSEvTime[0] = Double.NaN;
                unitSEvTime[1] = e.t;
                unitSEv = new Event(data.units[x][y],unitSEvTime,e.type);
              } else {
                unitSEv = e;
              }
              //No change in X,Y,Z in unit
              unitSEv.X = lastEventInUnit[x][y].X;
              unitSEv.Y = lastEventInUnit[x][y].Y;
              unitSEv.Z = lastEventInUnit[x][y].Z;
              //Link it up in the chain.
              unitSEv.prevEventInSameUnit = lastEventInUnit[x][y];
              unitSEv.prevEventInSameUnit.nextEventInSameUnit = unitSEv;
              lastEventInUnit[x][y] = unitSEv;
            }
          }
          break;
      } //end of switch
    } //end of for all events

    if (LadyBug.DEBUG) {
      System.out.println("<<<<<<<<<<<<<<<<<<<<");
      System.out.println(toStringEvents());
      System.out.println(">>>>>>>>>>>>>>>>>>");
    }
  }

  /**
   * Function to be called when initializing the theta0
   * and typicalParamSize initialization in the ML method. To be specialized
   * int subclasses
   *
   * @param theta0 Array of theta0's, i.e. the start values
   * @param typicalParamSize Typical size
   */
  public void MLParamsSetup(double[] theta0, double typicalParamSize[]) {
    //Beta and betaN
    theta0[beta_] = 1;
    if (data.noOfUnits>1) {theta0[betaN_] = 1;} else {theta0[betaN_] = 0; }
    //Infectious period
    if (Options.expInfTime) {
      theta0[gammaI_] = Math.log(1); //gammaI = Math.exp(theta0) and similar for deltaI.
    } else {
      theta0[gammaI_] = Math.log(0.1);
    }
    theta0[deltaI_] = Math.log(0.75);
    theta0[gammaE_] = Math.log(1);
    theta0[deltaE_] = Math.log(1);

    typicalParamSize[beta_] = 1;
    typicalParamSize[betaN_] = 1;
    typicalParamSize[gammaE_] = 0.1;
    typicalParamSize[deltaE_] = 0.75;
    typicalParamSize[gammaI_] = 0.1;
    typicalParamSize[deltaI_] = 0.75;
  }

  /**
   * Hook function called inside ML once the results are showns
   *
   * @param theta Array with the optimized valued. Index 1-4 are taken care of.
   */
  public void MLParamsPrintHook(double[] theta) {
    //Do nothing.
  }
  /**
   * Perform maximum likelihood estimation by numerically maximizing
   * the likelihood equation. In special cases some params can be derived
   * analytically.
   *
   * @param iHessian Inverse of Hessian as array (has to be newed before!)
   * @returns An array containing the MLE.
   */
  public double[] ML(double[][] iHessian) {
    System.out.println("--- ML  ---");

    //Turn the use of origins off
    useOrigins = false;
    //useOrigins = true;

    //Just make sure all structures etc. are initialized. Do not sample censored values.
    sampleState(false);
    genXYFuncs();

    ////////////////////////////////////////////////////////
    // Optimize beta, betaN, gammaI, deltaI and (possibly others), simultaneously
    ///////////////////////////////////////////////////////
    int n = noOfModelParams;
    double theta0[] = new double[n+1];
    double typicalParamSize[] = new double[n+1];
    double theta[]  = new double[n+1];
    double f[]  = new double[n+1];
    double g[]  = new double[n+1];
    int info[] = new int[2];
    double hessianWork[][] = new double[n+1][n+1];
    double udiag[] = new double[n+1];

    double gradtl[] = new double[2]; gradtl[1] = 1e-10;
    double stepmx[] = new double[2]; stepmx[1] = 20; //1 400
    double steptl[] = new double[2]; steptl[1] = 1e-10; //1e-10;
    double dlt[] = new double[2]; dlt[1] = 1e-6;
    int terminationCode[] = new int[2];
    int noOfDigits[] = new int[2]; noOfDigits[1] = 10;
    int maxNoOfIteration[] = new int[2]; maxNoOfIteration[1] = 150;
    int fExpensiveToEval[] = new int[2]; fExpensiveToEval[1] = 1;
    int method[] = new int[2]; method[1] = 1;
    double fscale[] = new double[2]; fscale[1] = 1;
    int gradientProvided[] = new int[2]; gradientProvided[1] = 0;
    int hessianProvided[]  = new int[2]; hessianProvided[1] = 0;

    //Call hook function to initialize extra parameters in subclasses.
    MLParamsSetup(theta0,typicalParamSize);

    //Show output of optimizer, y/n?
    Uncmin_f77.VERBOSE = false;
    //Optimize.
    //Uncmin_f77.optif0_f77(n,theta0,this,theta,f,g,info,hessianWork,udiag);

    //To optimize we operate with log transformed params
    expTransformParams = true;

    Uncmin_f77.optif9_f77(n,theta0, this,
                          typicalParamSize,fscale, method,
                          fExpensiveToEval, info,noOfDigits, maxNoOfIteration,
                          gradientProvided, hessianProvided, dlt,
                          gradtl,  stepmx, steptl,
                          theta, f, g,
                          terminationCode, hessianWork, udiag);

    //Assign params so we can use them
    for (int i=1; i<=n; i++) {
    System.out.println("theta["+i+"]=" + theta[i]);
    }
    
    array2Params(theta,expTransformParams);

    //Go back to original loglik mode!!
    expTransformParams = false;
    theta = params2Array();

    //Set representation
    form.setMinimumFractionDigits(3); //4
    form.setMaximumFractionDigits(3); //4

    //Compute Hessian to make sure we got the right thing.
    double a[][] = new double[n+1][n+1]; //the hessiaN
    double sx[] = new double[n+1];
    for (int i=1; i<=n; i++) { sx[i]=1;} //What is this?
    double wrk1[] = new double[n+1];
    double wrk2[] = new double[n+1];
    double epsm = 1.12e-16;
    double rnf = Math.max(Math.pow(10.0,-noOfDigits[1]),epsm);
    double fpls[] = new double[2];

    //Evaluate at empiral found value. Note that at this point expTransformParams
    //is turned to false, i.e. we operate on the REAL loglik function.
    fpls[1] = this.f_to_minimize(theta);

    System.out.println("-loglik(mle) = " + fpls[1]);
    //Compute the Hessian of -loglik (inverse of info matrix)
    Uncmin_f77.sndofd_f77(n,theta,this,fpls,a,sx,rnf,wrk1,wrk2);

    System.out.println("Inverse of Obs. Info Matrix = Hessian(-loglik): ");
    double diagmx = -1e99;
    for (int i=1; i<=n; i++) {
      for (int j=1; j<=n; j++) {
        //Find max of diagonal
        if ((i==j) && (a[i][j]>diagmx)) {diagmx = a[i][j];}
        //Complete the lower triangular matrix to the upper triangular
        if (i<j) { a[i][j] = a[j][i]; }
        System.out.print(form.format(a[i][j]) + "\t");
      }
      System.out.println();
    }

    //////////////////////////////////////////////////////////////////////
    // Diagnostic hack to check if likelihood is correct.
    ///////////////////////////////////////////////////////////////////////
    /*int gammaD_ = 7;
    int deltaD_ = 8;
    System.out.println("H[gammaD][gammaD] = " + a[gammaD_][gammaD_]);
    System.out.println("H[deltaD][deltaD] = " + a[deltaD_][deltaD_]);
    System.out.println("H[gammaD][deltaD] = " + a[gammaD_][deltaD_]);
    System.out.println("H[deltaD][gammaD] = " + a[deltaD_][gammaD_]);

    justDiagnostic = true;

    //Evaluate at empiral found value. Note that at this point expTransformParams
    //is turned to false, i.e. we operate on the REAL loglik function.
    //Note also that we now just have diagnostic times in the likelihood.
    fpls[1] = this.f_to_minimize(theta);
    System.out.println("Just diagnostic times: ");
    System.out.println("-loglik(mle) = " + fpls[1]);
    //Compute the Hessian of -loglik (inverse of info matrix)
    Uncmin_f77.sndofd_f77(n,theta,this,fpls,a,sx,rnf,wrk1,wrk2);

    System.out.println("H[gammaD][gammaD] = " + a[gammaD_][gammaD_]);
      System.out.println("H[deltaD][deltaD] = " + a[deltaD_][deltaD_]);
      System.out.println("H[gammaD][deltaD] = " + a[gammaD_][deltaD_]);
      System.out.println("H[deltaD][gammaD] = " + a[deltaD_][gammaD_]);
*/
    //////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////////
    int gammaD_ = 7;
    int deltaD_ = 8;
    System.out.println("H[gammaD][gammaD] = " + a[gammaD_][gammaD_]);
    System.out.println("H[deltaD][deltaD] = " + a[deltaD_][deltaD_]);
    System.out.println("H[gammaD][deltaD] = " + a[gammaD_][deltaD_]);
    System.out.println("H[deltaD][gammaD] = " + a[deltaD_][gammaD_]);

    //Find cholesky decomposition of H to use solve function Ax=b
    double addmax[] = new double[n+1];
    double tol = 1e-10;
    Uncmin_f77.choldc_f77(n, a, diagmx, tol, addmax);

    //Allocate the inverse Hessian matrix to be returned
    //iHessian = new double[n+1][n+1];

    //Compute the inverse of the Hessian by using the Ax=b solution method
    double b[] = new double[n+1];
    double x[] = new double[n+1];
    //Coz it can only handle x and b to be of vector-type we manually need to loop
    //over all column vectors in A^-1 and I as in A * A^-1 = I.
    for (int j=1; j<=n; j++) {
      //Initialize the b vector as the j'th column of the identity matrix
      for (int i=1; i<=n; i++) {
        b[i] = (i == j) ? 1 : 0;
      }

      //Solve the equation system (one column of H^-1 at the time)
      Uncmin_f77.lltslv_f77(n, a, x, b);

      //Copy of the result
      for (int i=1; i<=n;  i++) { iHessian[i][j] = x[i]; }
    }

    System.out.println("\n\nCovariance Matrix = Hessian^{-1}[loglik]");
    for (int i=1; i<=n; i++) {
      for (int j=1; j<=n; j++) {
        System.out.print(form.format(iHessian[i][j]) + " ");
      }
      System.out.println();
    }


    System.out.println("Cov[gammaD][gammaD] = " + iHessian[gammaD_][gammaD_]);
    System.out.println("Cov[deltaD][deltaD] = " + iHessian[deltaD_][deltaD_]);
    System.out.println("Cov[gammaD][deltaD] = " + iHessian[gammaD_][deltaD_]);
    System.out.println("Cov[deltaD][gammaD] = " + iHessian[deltaD_][gammaD_]);

    //Show the result together
    System.out.println("--------------------------------------");
    System.out.println("\nParameters: (\\hat,s.e.(\\hat)) ");
    for (int i=1; i<=this.noOfModelParams; i++) {
      System.out.println(this.modelParamNames[i] + "\t= (" + theta[i] + "\t, " + Math.sqrt(iHessian[i][i]) + ")");
    }

    System.out.println("\nRelevant correlations: ");
    for (int i=1; i<=this.noOfModelParams; i=i+2) {
      System.out.println("Corr(" + this.modelParamNames[i] + "," + this.modelParamNames[i+1] + ")=\t" + iHessian[i][i+1]/Math.sqrt(iHessian[i][i]*iHessian[i+1][i+1]));
    }

    //Print additional parameters as well.
    System.out.println("\n" + toStringDiagnostic());

    //Return -- note that also the inverse Hessian is returned by call
    //by reference.
    return(theta);
  }
  /** Foo function to be specialized in SEIRWState */
  public void MLProfiles() {
    //Do nada.
  }


  /** Count the number of exposures originating from the unit itself.
   *
   * @return E_{loc}
   */
  public LinkedList Eloc() {
    LinkedList res = new LinkedList();

    //Easier to loop over all events than E, coz then we would have to take
    //the e_1 problem into account. Note we start at one to not inclue e_1.
    for (int i=1; i<noOfEvents; i++) {
      Event e = events[i];
      //Only interested in exposed events.
      if ((e.type == 'E') && (e.origin == e.unit)) {
        res.add(e);
      }
    }
    return(res);
  }
  /** Count the number of exposures originating from a neighbour.
   *
   * @return E_{nbr}
   */
  public LinkedList Enbr() {
    LinkedList res = new LinkedList();
    //Easier to loop over all events than E, coz then we would have to take
    //the e_0 problem into account.
    for (int i=1; i<noOfEvents; i++) {
      Event e = events[i];
      //Only interested in exposed events.
      if ((e.type == 'E') && (e.unit.isNeighbour(e.origin))) {
        res.add(e);
      }
    }
    return(res);
  }

  /** Hook function to check for a valid state. In our case it's a special case if its
   * a disease transmission experiment. Then all event times have to be in >= 0
   *
   * @return Boolean indicating whether all event times are >=0.
   */
  public boolean isValidHook() {
    if (!Double.isNaN(data.Ttime)) {
      for (int i=0;i<data.noOfIndividuals; i++) {
        if ((E[i].t<0) || (I[i].t<0) || (R[i].t<0)) {
          return(false);
        }
      }
    }
    return(true);
  }

  /** Hook function for header writing */
  public String toHeaderHook() { return("");}
 
  /** Hook function for SEIR specific output */
  public String toStringHook() {
    StringBuffer s = new StringBuffer();
    
    for (int i=0; i<data.noOfIndividuals; i++) {
      s.append("\t" + form.format(I[i].t - E[i].t));
    }
    for (int i=0; i<noOfIParam; i++) {
      s.append("\t" + form.format(I[IParamIndex[i]].t));
    }
    return(s.toString());
  }


  public String showStats(int noOfIterations) {
    StringBuffer s = new StringBuffer();

    NumberFormat nf = new DecimalFormat("00.00%");

    //Show all acceptance rates
    s.append("\nE acceptance rate   : " + nf.format((double)Eaccept/(double)(runs)) + "\n");
    s.append("I acceptance rate   : " + nf.format((double)Iaccept/(double)(runs)) + "\n");
    s.append("R acceptance rate   : " + nf.format((double)Raccept/(double)(runs)) + "\n");
    for (int i=1;i<=this.noOfModelParams;i++) {
      s.append(this.modelParamNames[i] + " acceptance rate : " + nf.format((double)accept[i]/(double)(runs)) + "\n");
    }
    s.append("Individual event acceptance rates:\n");
    Iterator it = paramList.iterator();
    int i=0;
    while (it.hasNext() && (paramList.size()>0)) {
      Event e = (Event)it.next();
      s.append(e.type + ":" + nf.format(e.accepted/((double)(runs))) + "\n");
    }

    //Open logfile and show stats
    s.append(analLog("log.txt"));

    //Return the result.
    return(s.toString());
  }

  /** Constructor - new necessary structures*/
  public SEIRState() {
    //Call superclass constructor.
    super();

    //Find out how much is not known about E values.
    noOfEParam = data.getNoOfUnknownE();
    EParamIndex = data.getUnknownEIndexes();


    //Number of model parameters depends on whether incubation time is a fixed constant.
    /*if (Options.incuTime == 'C') {
    //Model parameters are beta,betaN,gammaI,deltaI
    //Note to be compatible with optimizer we use Fortran array starting at 1.
      noOfModelParams = 4;
      modelParamNames = new String[] {"FORTRAN","beta","betaN","gammaI","deltaI"};
    //Indices
      beta_ = 1; betaN_ = 2; gammaI_ = 3; deltaI_ = 4;
    } else {*/
    //Model parameters are beta,betaN,gammaI,deltaI.
    noOfModelParams = 6;
    modelParamNames = new String[] {"FORTRAN","beta","betaN","gammaE","deltaE","gammaI","deltaI"};
    beta_ = 1; betaN_ = 2; gammaE_ = 3; deltaE_ = 4; gammaI_ = 5; deltaI_ = 6;
    //}
    theta = new double[noOfModelParams+1];
    accept = new int[noOfModelParams+1];

    //Which parametrisation are we using. This should at some point go into the options file.
    MeanVarianceParameters = Options.meanVar;

    //Generate the speedup structure for lastEventInit
    lastEventInUnits  = new Event[data.noOfUnitsX+1][data.noOfUnitsY+1];
    firstEventInUnits = new Event[data.noOfUnitsX+1][data.noOfUnitsY+1];
  }
} //End of class SEIRState