package sir.model;

import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.io.*;
import java.util.*;
import org.omegahat.Simulation.MCMC.*;

//My own stuff.
import sir.optim.*;
import sir.util.*;
import sir.estimate.*;

/**
 * Subclass implementing all functionality for an multivariate SEIRwD model.
 * 
 * Extra functions for handling diagnostic events is added. Most is done
 * using hook functions. However, the ML function is coded pretty much from scratch.
 *
 * @author Michael Hhle &lt;hoehle@stat.uni-muenchen.de&gt;
 * @version 5 January 2003
 *
 */
public class SEIRwDState extends SEIRState implements MCMCState,Uncmin_methods {
  /** Parameters of the gamma distributed seroconversion time */
  public double gammaD, deltaD;
  /** Indices in param array */
  public int gammaD_,deltaD_;

  /** All seroconversion events */
  public Event[] D;
  /** Array of indexes showing which recovery times are unknown and hence are variables */
  int DParamIndex[];
  /** Number of parametric S */
  int noOfDParam;

  /** Clone. Not everything is cloned. Just enough to be able to reconstruct likelihood
   * 
   */
  public Object clone() {
    SEIRwDState clone = new SEIRwDState();
    
    //Init number formatter
    form = LadyBug.form;

    //Fetch interesting numbers from the Data class
    clone.noOfEParam = noOfEParam;
    clone.noOfIParam = noOfIParam;
    clone.noOfRParam = noOfRParam;
    clone.noOfDParam = noOfDParam;
    
    //The next two arrays are not altered so we borrow the reference. If changed
    //it will change here as well!!
    clone.IParamIndex = IParamIndex; 
    clone.RParamIndex = RParamIndex;
    clone.EParamIndex = EParamIndex;
    clone.DParamIndex = DParamIndex;
    
    //Model parameters - just borrow the array structure
    clone.noOfModelParams = noOfModelParams;
    clone.modelParamNames = modelParamNames;
    
    //Names to index
    clone.beta_ = beta_;
    clone.betaN = betaN;
    clone.gammaE_ = gammaE_;
    clone.deltaE_ = deltaE_; 
    clone.gammaI_ = gammaI_; 
    clone.deltaI_ = deltaI_;
    clone.gammaD_ = gammaD_;
    clone.deltaD_ = deltaD_;
    
    //Everything to do with parameters we better clone
    accept = new int[noOfModelParams+1];
    theta = new double[noOfModelParams+1];

    //Clone the theta vector
    clone.theta = this.params2Array();
    clone.array2Params(clone.theta,false);

    //Clone all events - this is memory & time consuming!!.
    clone.E = new Event[data.noOfIndividuals];
    clone.I = new Event[data.noOfIndividuals];
    clone.R = new Event[data.noOfIndividuals];
    clone.D = new Event[data.noOfIndividuals];
   
    for (int i=0; i<data.noOfIndividuals; i++) {
      clone.E[i] = (Event)E[i].clone();
      clone.I[i] = (Event)I[i].clone();
      clone.R[i] = (Event)R[i].clone();
      clone.D[i] = (Event)D[i].clone();
    }

    //Use origins
    clone.useOrigins = useOrigins;

    //Generate the speedup structure for lastEventInit
    clone.lastEventInUnits  = new Event[data.noOfUnitsX+1][data.noOfUnitsY+1];
    clone.firstEventInUnits = new Event[data.noOfUnitsX+1][data.noOfUnitsY+1];

    //Which parametrisation are we using. This should at some point go into the options file.
    clone.MeanVarianceParameters = MeanVarianceParameters;
    
    
    //Generate the XYZ functions
    clone.genXYFuncs();
    
    //Now we are finally done!!
    return(clone);
  }
  
  /**
   * Transform from mean,variance to gammaX,deltaX for all parameters, where
   * this applies. Reverse it by calling gammaDelta2MeanVar.
   * @see gammaDelta2MeanVar
   */
  public void meanVar2GammaDelta() {
    //Translate from (Imean, Ivariance) to (gammaI,sigma). For convienience
    //Imean and Ivar are just gammaI and deltaI until now, where we will
    //store the real values and retrieve once them loglik has been calc'ed.
    double Imean = gammaI;
    double Ivar  = deltaI;
    double Dmean = gammaD;
    double Dvar  = deltaD;

    //Back transform x = r^-1(y)
    gammaI = (Imean*Imean) / Ivar;
    deltaI = Imean / Ivar;
    gammaD = (Dmean*Dmean) / Dvar;
    deltaD = Dmean / Dvar;
  }
  /**
   * The reverse of @see meanVar2GammaDelta, i.e. y=r(x)
   */
  public void gammaDelta2MeanVar() {
    //Forward transform y = r(x)
    double Imean = gammaI/deltaI;
    double Ivar  = gammaI/(deltaI*deltaI);
    double Dmean = gammaD/deltaD;
    double Dvar  = gammaD/(deltaD*deltaD);

    //We operate only on one set of parameters, so this is is just
    //the moving of values. Note that r(r^-1(y)) = y
    gammaI = Imean;
    deltaI = Ivar;
    gammaD = Dmean;
    deltaD = Dvar;

  }


  /** Hook function returning how many additional events are put into the
   * events array by subclasses etc. Overwrites similar method in SEIRState.
   *
   * @return Number of additional events added.
   */
  protected int extraNoOfEventsHook() {
    return data.noOfIndividuals;
  }

  /** Hook function to add additional events to the events array.
   *  Overwrites method in SEIRState class.
   *
   * @param baseNoOfEvents Number of E,I,R events in SEIRState class.
   * @return Number of additional events added.
   */
  protected void addExtraEventsHook(int baseNoOfEvents) {
    System.arraycopy(this.D,0,events,baseNoOfEvents,extraNoOfEventsHook());
  }

  /** Hook function for adding additional events to the loglikelihood. It is called
   * from within loglikGammaDelta. The rec-inf periods are already added.
   * In this case it is the (possibly censored) seroconversion events.
   * Remember that we operate on log scale!!
   *
   * @param i Number of the individual.
   * @return Likelihood of the additional events.
   */
  protected double loglikAddIndividualHook(int i) {
    //System.out.println("#"+ i +": e=" + (D[i].t));
    //System.out.println("#"+ i +": d-e=" + (D[i].t-E[i].t) + "\tparametric? " + D[i].isParametric + "\tcensored? " + D[i].isCensored);
    
    //For MCMC context hack -- this stinks.
    //if (true) {
    //  return(prob.loggammaPDF(D[i].t - E[i].t,gammaD,deltaD));
    //}

    //Likelihood of the seroconversion events depends on whether censoring has occured.
    //If it is not censored or parametric because we are sampling its values then
    //use the direct form. Otherwise it's the Survival function we need.
    if (!D[i].isCensored || D[i].isParametric) { //not censored - just log pdf
      //System.out.println("#"+ i +": d-e=" + (D[i].t-E[i].t) + "\tparametric? " + D[i].isParametric + "\tcensored? " + D[i].isCensored);
    
      return(prob.loggammaPDF(D[i].t - E[i].t,gammaD,deltaD));
    } else {
      //censored -- survival func of censoring variable (S[i].censorTime contains the value!)
      //Can't use S[i].t, coz we probably sampled a value for it.
      //System.out.println("gammaD=" + gammaD);
      //System.out.println("deltaD=" + deltaD);
      //System.out.println("s - e =" + (S[i].t - E[i].t));
      double S = prob.gammaSurv(D[i].censorTime - E[i].t,gammaD,deltaD);
      //Handle that if gammaD,deltaD are bad then we could potentially obtain a zero
      //making the likelihood VERY negative ;-)
      //System.out.println("#"+ i +": d-e=" + (D[i].censorTime-E[i].t) + "\tparametric? " + D[i].isParametric + "\tcensored? " + D[i].isCensored);
      if (S == 0) {
        System.out.println("S is zero!!");
        return(-1e99);
      } else {
        return(Math.log(S));
      }
      /*
      System.out.println("D[i].censorTime=" + D[i].censorTime);
      System.out.println("D[i].t         =" + D[i].t);
      System.out.println("E[i].t         =" + E[i].t);
      System.out.println("D.censor-E.t   =" + (D[i].censorTime - E[i].t));
      System.out.println("gammaD=" + gammaD + "\tdeltaD=" + deltaD);
      System.out.println("Surv           =" + prob.gammaSurv(D[i].censorTime - E[i].t,gammaD,deltaD));
      System.out.println("S(s-e) = " + res);
      */
    }
  }


  /**
   * Hook function to be called when initializing the theta0
   * and typicalParamSize initialization in the ML method. Add seroconversion part.
   *
   * @param theta0 Array of theta0's, i.e. the start values
   * @param typicalParamSize Typical size
   */
  public void MLParamsSetup(double[] theta0, double typicalParamSize[]) {
    super.MLParamsSetup(theta0,typicalParamSize);

    //Now our stuff
    theta0[gammaD_] = Math.log(1);
    theta0[deltaD_] = Math.log(1);
    typicalParamSize[gammaD_] = Math.log(1);
    typicalParamSize[deltaD_] = Math.log(1);
  }

  /** Create Profile Plots of LogLikelihood around the max */
  public void MLProfiles() {
    int p1_=0, p2_=0;
    String profileFileName = null;

    for (int i=0; i<2; i++) {
      switch (i) {
      case 0:
        p1_ = beta_;
        p2_ = betaN_;
        profileFileName = "betabetaNprofile.txt";
        break;
      case 1:
        p1_ = gammaD_;
        p2_ = deltaD_;
        profileFileName = "gammaDdeltaDprofile.txt";
        break;
      case 2:
        p1_ = gammaE_;
        p2_ = deltaE_;
        profileFileName = "gammaEdeltaEprofile.txt";
        break;
      case 3:
        p1_ = gammaI_;
        p2_ = deltaI_;
        profileFileName = "gammaIdeltaIprofile.txt";
        break;
    }

    //Fetch ML estimate
    double iHessian[][] = new double[noOfModelParams+1][noOfModelParams+1];
    double theta[] = ML(iHessian);
    double p1 = theta[p1_];
    double p2 = theta[p2_];
    double logmax = -this.f_to_minimize(theta);
    //Fetch standard deviations.
    double sep1 = Math.sqrt(iHessian[p1_][p1_]);
    double sep2 = Math.sqrt(iHessian[p2_][p2_]);

    /*double plusminus = 0.25;
    double p1Step = p1*plusminus/100.0;
    double p2Step = p2*plusminus/100.0;
    double p1Min = p1*(1-plusminus);
    double p2Min= p2*(1-plusminus);
    double p1Max = p1*(1+plusminus);
    double p2Max = p2*(1+plusminus);*/


    double p1Min = p1-2.5*sep1;
    double p2Min= p2-2.5*sep2;
    double p1Max = p1+2.5*sep1;
    double p2Max = p2+2.5*sep2;

    double noOfEvals = 51;
    double p1Step = (p1Max-p1Min)/noOfEvals;
    double p2Step = (p2Max-p2Min)/noOfEvals;

    //Ensure we do not calculate with exp'ed values.
    expTransformParams = false;

    //Create log file txt
    try {
      PrintStream txt = new PrintStream(new FileOutputStream(profileFileName));
      //First line contains a header of the form:
      //p1max    s.e.(p1)    p2max s.e.(p2)   cov(p1,p2)    logmax
      txt.println(p1 + "\t" + sep1 + "\t" + p2 + "\t" + sep2 + "\t" + iHessian[p1_][p2_] + "\t" + logmax);
      txt.println("p1\tp2\tloglik");
      theta[p1_]  =  p1Min;
      while (theta[p1_] < p1Max) {
        theta[p2_] =  p2Min;
        while (theta[p2_] < p2Max) {
          //Calculate likelihood
          double loglik = - f_to_minimize(theta);

          //Show result
          txt.println(theta[p1_] + "\t" + theta[p2_] + "\t" + loglik);

          //Update p2
          theta[p2_] += p2Step;
        }
        //Update p1
        theta[p1_] += p1Step;
      }
      //End of profile calculations.
      txt.close();
    } catch (FileNotFoundException e) {
    }
  }
  }



  /**
   * Function to display how parameters relate to data.
   *
   */
  public String toStringDiagnostic() {
    //Call super method
    StringBuffer s = new StringBuffer(super.toStringDiagnostic());

    //Calculate data mean of seroconversions
    double sum=0;
    for (int i=0; i<data.noOfIndividuals; i++) {
      sum += D[i].t-E[i].t;
    }
    double mean = sum/data.noOfIndividuals;
    sum = 0;
    for (int i=0; i<data.noOfIndividuals; i++) {
      sum += Math.pow((D[i].t - E[i].t) - mean,2);
    }
    double var = sum/(double)(data.noOfIndividuals-1);

    s.append("Data mean of seroconversions= (" + LadyBug.form.format(mean) + "\t, " + LadyBug.form.format(var) +")\n");
    if (MeanVarianceParameters) {
      s.append("Estimated Mean     =" + LadyBug.form.format(gammaD) + "\tVariance = " + LadyBug.form.format(deltaD) + ")\n");
    } else {
      s.append("Estimated mean              = (" + LadyBug.form.format(gammaD/deltaD) + "\t, " +  LadyBug.form.format(gammaD/Math.pow(deltaD,2)) + ")\n");
    }

    return(s.toString());
  }

  /** Helper function, given an array of size noOfModelParams assign
   * the values of this vector to the correct names. Call supermethod and initializae gammaD,deltaD
   *
   * @param theta - A vector of length noOfModelParams containing the values to assign
   * @param expTransformParams - When using the minimizer it is a good idea to operate with
   *                             log(param) value instead to avoid negative values.
   */
  public void array2Params(double theta[], boolean expTransformParams) {
    //Call super method
    super.array2Params(theta,expTransformParams);

    //The extra stuff
    deltaD = expTransformParams ? Math.exp(theta[deltaD_]) : theta[deltaD_];
    if (Options.diagTime == 'E') {
    gammaD = expTransformParams ? Math.exp(1) : 1;
    } else {
      gammaD = expTransformParams ? Math.exp(theta[gammaD_]) : theta[gammaD_];
    }
  }

  /** Helper functions to put all parameters in an array. Remember Fortran indexing
    * Function currently does not work if incu period a constant.
    *
    * @return An @see noOfModelParams + 1 long array with params
    */
   public double[] params2Array() {
     double theta[] = super.params2Array();

     //The new ones
     theta[gammaD_] = gammaD;
     theta[deltaD_] = deltaD;

     return(theta);

  }
  /** This version shows only time of exposure and the length of the infectious time */
  public String toStringEvents2() {
    StringBuffer s = new StringBuffer("\n");
    for (int i=0; i<data.noOfIndividuals; i++) {
      s.append(i + ":\t" + LadyBug.form.format(E[i].t) + "\t" + LadyBug.form.format(R[i].t-I[i].t)+"\n");
    }
    return(s.toString());
  }
  
  /** Show all stats */
  public String toStringEvents3() {
    StringBuffer s = new StringBuffer("\n");
    s.append("i\tE\tI\tD\n");
    for (int i=0; i<data.noOfIndividuals; i++) {
      s.append(i + ":\t" + LadyBug.form.format(E[i].t) + "\t" + LadyBug.form.format(I[i].t) + "\t" + LadyBug.form.format(D[i].t) + "\n");
    }
    return(s.toString());
  }


  /** Hook function to check for a valid state. In our case it's a special case if its
   * a disease transmission experiment. Then all event times have to be in >= 0
   *
   * @return Boolean indicating whether all event times are >=0.
   */
  public boolean isValidHook() {
    if (Double.isInfinite(data.Ttime)) {
      for (int i=0;i<data.noOfIndividuals; i++) {
        if ((E[i].t<0) || (I[i].t<0) || (R[i].t<0)) {
          return(false);
        }
      }
    }
    return(true);
  }

  /** Hook function for header writing */
  public String toHeaderHook() {
    StringBuffer s = new StringBuffer();
    //Try to convert object to SEIRwDState object. If it works then write
    //the rest. A better way would be to put this into each)
    SEIRwDState sswd = ((SEIRwDState)this);
    for (int i=0; i<sswd.noOfDParam; i++) {
      s.append("\td" + sswd.DParamIndex[i]);
    }
    return(s.toString());
  }         
  /** Hook function calling the super before adding SEIRwD specifc things*/
  public String toStringHook() {
    StringBuffer s = new StringBuffer(super.toStringHook());
    for (int i=0; i<noOfDParam; i++) {
      s.append("\t" + form.format(D[DParamIndex[i]].t));
    }
    return(s.toString());
  }
 
  /**
   *  Initializing gammaD and deltaD is done using this hook function.
   */
  protected void sampleStateParamsHook() {
    //Extra initialization - new the diag array
     D = new Event[data.noOfIndividuals];
  }

  /**
   * Use hook function to new individual's diagnostic event in case of censoring.
   * We currently only handle censored diagnostic events, not completly missing.
   *
   * @param i Number of the individual we are newing events for.
   */
  protected void sampleStateEventsHook(int i, boolean sampleCensored) {
    //Copy the event to the state
    D[i] = data.D[i];
    //Sample a time if it is parametric
    if (D[i].isParametric) {
      D[i].t = E[i].t + prob.gammaTruncRand(gammaD,deltaD,D[i].censorTime-E[i].t);
    }
    //But if it is censored and we do not want sampled event time then
    //we fetch the censor value
    if ((D[i].isCensored) && (!sampleCensored)) {
      D[i].t = D[i].censorTime;
      D[i].isParametric = false;
    }

    //If it is still parametric then we add it. But they are not really interesting
    //as Gibbs sampling is used.
    //if (D[i].isParametric) {
    //  paramList.add(D[i]);
    //}
  }

  /** Update gammaD using Metropolis Hastings. This new version includes only relevant
   * parts of the likelihood when calculating the acceptance prob. Quite a speed gain!
   * Note though mean/var representation does not work now, but it sucks anyway.
   */
  public void updateGammaD() {
    //If exponential distribution we do not need to do anything.
    if (Options.expInfTime) {gammaD = 1; return;}
    //Safe but slow Metropolis Hastings function
    /*if (Options.slowButSafe) {
      double theta[] = this.params2Array();
      updateMetroHastings(theta,gammaD_, Options.mcmc_gammaDRWsigma);
      return;
    }*/


    //Calculate log Y = \sum_{j\in A_T} log(d_j-e_j)
    double logY = 0;
    for (int i=0; i< data.noOfIndividuals; i++) {
      logY += Math.log(D[i].t - E[i].t);
    }

    //Save values
    double savegammaD = gammaD;

    //Find value of proportional likelihood part evaluated at old value
    //instead of just using: loglikold  = loglik();
    double loglikold  = data.noOfIndividuals * gammaD * Math.log(deltaD) + (gammaD-1)*logY - data.noOfIndividuals * prob.lngamma(gammaD);
    // Posterior evaluated at old value: lik \times prior
    double logfold = loglikold + (data.nu[gammaD_]-1) * Math.log(gammaD) - data.lambda[gammaD_]*gammaD;


    // Make proposal value gammaD'_{r} - use abs(val) - property of Gaussian distribution
    gammaD = Math.abs(gammaD + prob.normalRand()*Options.mcmc_gammaDRWsigma);

    //Check new loglik - f(R,I-gammaD_{xyr}+gammaD_{xyr}'). Again we only evaluate relevant parts.
    double logliknew  = data.noOfIndividuals * gammaD * Math.log(deltaD) + (gammaD-1)*logY - data.noOfIndividuals * prob.lngamma(gammaD);
    // Posterior evaluated at new value: lik \times prior
    double logfnew = logliknew + (data.nu[gammaD_]-1) * Math.log(gammaD) - data.lambda[gammaD_]*gammaD;

    //Accept the new value y/n?
    if (!acceptProposal(logfnew,logfold)) {
      gammaD = savegammaD;
    } else { //accept
      accept[gammaD_]++;
    }
  }


  /**
   * Gibbs updating of deltaD - we know the full conditional (see paper)
   */
  public void updateDeltaD() {
    /*if (Options.slowButSafe) {
      double theta[] = this.params2Array();
      updateMetroHastings(theta,deltaD_, Options.mcmc_deltaDRWsigma);
      return;
    }*/

    // a = N*gammaD + nu[deltaD_]
    double a = data.noOfIndividuals * gammaD + data.nu[deltaD_];
    // b = \sum_{i=1}^N (s_j - e_j) + \lambda[deltaD_]
    double b = data.lambda[deltaD_];
    for (int i=0; i<data.noOfIndividuals; i++) {
      b += (D[i].t - E[i].t);
    }
    deltaD = prob.gammaRand(a,b);
    //GibbsUpdating always accepts
    accept[deltaD_]++;
  }

 /** Update E using full conditionals. Currently, Metropolis Hastings is used, but
  * only relevant parts of likelihood are computed.
  * This version includes Diagnostic events.
  *
  */
 public void updateE() {
   //Loop over all E in EParamIndex to updateIn case E is fixed nothing needs to be done
   //Make sure e1 is not parametric!
   int batchAccept = 0;

   //Loop over all E events who need to be updated
   for (int m=0; m<this.noOfEParam; m++) {
     //Find the index we want.
     int j = this.EParamIndex[m];
     //Save its stats - i.e. time and current likelihood*prior
     double saveE = E[j].t;

     //For each we (unfortunately need to loop in the chain to the event
     //so the lastEvent structure is correct. Snapshots would be great here!


     /////////////////////////////////////////////////////////////////
     // Initialize data structure keeping track of the last event (le) in
     // each section.
     ////////////////////////////////////////////////////////////////
     Event le[][] = this.lastEventInit(true); //new array coz betaSurv uses it as well!!!

     //Loop correctly updating le until we find the event of interest.
     int i=0;
     while (events[i] != E[j]) {
       Event e = events[i];
       le[e.posx][e.posy] = e;
       i++;
     }


     //Calculate w(X) = pi(x)/q(X) according to p.9 in the Gilks book. Note: Proposal q(X)
     //is of independence sampler type, hence we use eq. (1.8). Not symmetric!!
     double loglikOld = Math.log(betaHaz(E[j].unit,E[j].origin,le)) + betaSurvLn(E[j].unit,E[j].t);
     double logWOld = loglikOld + (gammaE-1)*Math.log(I[j].t-E[j].t) + (gammaD-1)*Math.log(D[j].t-E[j].t) - deltaE*(I[j].t-E[j].t) - deltaD*(D[j].t-E[j].t); ;

     //To confirm the code
     //double logWOld = loglik();

     //Propose new value, use q(), which is a good proposal density.
     //Coz we will reject all exposure times before e1=0 we loop
     //until we get a valid one. Is this valid?
     double foo;
     do {
       //We should have some sigma parameter open for tuning.
       //Do we get probs with the symmetric properties of E here?
       foo  = Math.abs(E[j].t + prob.normalRand()*Options.mcmc_ERWsigma);
     } while ((foo>=I[j].t) || (foo>=D[j].t) || (foo<=0));
     E[j].t = foo;

     //Independence sampler:
     //E[j].t = prob.uniformRand()*Math.min(I[j].t,D[j].t);

     //To make sure time order is still sorted, we better recompute all events.
     genXYFuncs();


     //Move up the list again to get right neighbour values.
     le = this.lastEventInit(true); //new array coz betaSurv uses it as well!!!
     //Loop correctly updating le until we find the event of interest.
     i=0;
     while (events[i] != E[j]) {
       Event e = events[i];
       le[e.posx][e.posy] = e;
       i++;
     }

     //Calculate w(Y) = pi(Y)/q(Y).
     double loglikNew = Math.log(betaHaz(E[j].unit,E[j].origin,le)) + betaSurvLn(E[j].unit,E[j].t);
     double logWNew = loglikNew + (gammaE-1)*Math.log(I[j].t-E[j].t) + (gammaD-1)*Math.log(D[j].t-E[j].t) - deltaE*(I[j].t-E[j].t) - deltaD*(D[j].t-E[j].t);


     //To confirm the code
     //double logWNew = loglik();

     //Check whether proposal is accepted and do appropriate action.
     if (!acceptProposal(logWNew,logWOld)) {
       //Restore the old value and restore all events (might be superfluous)
       E[j].t = saveE;
       genXYFuncs();
     } else {
       batchAccept++;
       E[j].accepted++;
     }
   }
   //System.out.println("batchAccept = " +  batchAccept);
   //Add to acceptance rates.
   Eaccept += batchAccept/(double)noOfEParam;
 }

 /** Update E using full conditionals. Currently, Metropolis Hastings is used, but
  * only relevant parts of likelihood are computed.
  * This version includes Diagnostic events.
  *
  */
 public void updateEPropagate() {
   //Loop over all E in EParamIndex to updateIn case E is fixed nothing needs to be done
   //Make sure e1 is not parametric!
   int batchAccept = 0;

   //Loop over all E events who need to be updated
   for (int m=0; m<this.noOfEParam; m++) {
     //Find the index we want.
     int j = this.EParamIndex[m];
     //Save its stats - i.e. time and current likelihood*prior
     double saveE = E[j].t;

     //For each we (unfortunately need to loop in the chain to the event
     //so the lastEvent structure is correct. Snapshots would be great here!


     //Calculate w(X) = pi(x)/q(X) according to p.9 in the Gilks book. Note: Proposal q(X)
     //is of independence sampler type, hence we use eq. (1.8). Not symmetric!!
     double logWOld =  (gammaE-1)*Math.log(I[j].t-E[j].t) + (gammaD-1)*Math.log(D[j].t-E[j].t) - deltaE*(I[j].t-E[j].t) - deltaD*(D[j].t-E[j].t); ;

     //To confirm the code
     //double logWOld = loglik();

     //Propose new value, use q(), which is a good proposal density.
     //Coz we will reject all exposure times before e1=0 we loop
     //until we get a valid one. Is this valid?
     double foo;
     do {
       //We should have some sigma parameter open for tuning.
       //Do we get probs with the symmetric properties of E here?
       foo  = Math.abs(E[j].t + prob.normalRand()*Options.mcmc_ERWsigma);
     } while ((foo>=I[j].t) || (foo>=D[j].t));
     E[j].t = foo;

     //To make sure time order is still sorted, we better recompute all events.
     genXYFuncs();

     double logWNew = (gammaE-1)*Math.log(I[j].t-E[j].t) + (gammaD-1)*Math.log(D[j].t-E[j].t) - deltaE*(I[j].t-E[j].t) - deltaD*(D[j].t-E[j].t);


     //To confirm the code
     //double logWNew = loglik();

     //Check whether proposal is accepted and do appropriate action.
     if (!acceptProposal(logWNew,logWOld)) {
       //Restore the old value and restore all events (might be superfluous)
       E[j].t = saveE;
       genXYFuncs();
     } else {
       batchAccept++;
       E[j].accepted++;
     }
   }
   //System.out.println("batchAccept = " +  batchAccept);
   //Add to acceptance rates.
   Eaccept += batchAccept/(double)noOfEParam;
 }

 /** Method to update the D. As shown in paper these are only the censored events
  * which can be updated using Gibbs sampling from a truncated gamma distribution */
 public void updateD() {
   //Only update the parametric (i.e. the censored ones)
   for (int i=0; i<this.noOfDParam; i++) {
     int j = DParamIndex[i];
     double TD = prob.gammaTruncRand(gammaD,deltaD,D[j].censorTime-E[j].t);
     //If the probability to achieve values greater than the truncation time
     //is extremely low  we could get NaN's instead of values.
     //Instead we will pick the censorTime as value.
     if (Double.isNaN(TD)) {
       System.err.println("Warning: TD is NaN for individual " + j + ", i.e. truncation time very unlikely.");
       System.err.println("gammaD=\t" + gammaD);
       System.err.println("deltaD=\t" + deltaD);
       System.err.println("D[j].censorTime =\t" + D[j].censorTime);
       System.err.println("E[j].t =\t" + E[j].t);
       System.err.println("f_{D-E}(" + (D[j].censorTime-E[j].t) + ")=\t" + prob.gammaPDF(gammaD,deltaD));
       //System.exit(-1);
       D[j].t = D[j].censorTime;
     } else {
       D[j].t = E[j].t + TD;
     }
   }

 }

  /**
   * Hook method to update additional parameters using MCMC, i.e. deltaD and gammaD
   * @param justTheta Just update theta no missing events.
   */
  public void updateHook(boolean justTheta) {
    updateGammaD();
    updateDeltaD();
    if (!justTheta) { updateD();}
  }


  /*public String showStats(int noOfIterations) {
    StringBuffer s = new StringBuffer(super.showStats(noOfIterations));
    NumberFormat nf = new DecimalFormat("00.00%");

    //Return the result.
    return(s.toString());
  }*/

  /** Constructor - new necessary structures*/
  public SEIRwDState() {
    //Call superclass constructur.
    super();

    //Find out how much is not known about S values, i.e. how many are censored.
    noOfDParam = data.getNoOfUnknownD();
    DParamIndex = data.getUnknownDIndexes();

    //Model parameters are superclass params (beta,betaN,gammaI,deltaI) and gammaD and deltaD
    noOfModelParams = 8;
    modelParamNames = new String[] {"FORTRAN","beta","betaN","gammaE","deltaE","gammaI","deltaI","gammaD","deltaD"};
    gammaD_ = 7;
    deltaD_ = 8;
    accept = new int[noOfModelParams+1];
    theta = new double[noOfModelParams+1];
  }

} //End of class SEIRState