package sir.sim;

import org.omegahat.Simulation.RandomGenerators.*;
import sir.util.*;
import java.text.NumberFormat;
import java.util.*;
import java.io.*;
import sir.estimate.*;

/**
   New sampler to handle Sellke constructions. Currently, only the single homogenous
   setup is handled. Intention is to go to a spatial setup at some point.
   Added seroconversion.

   * Added so vaccine effects etc. can calculate predictive posterior distributions.

   @author Michael Hhle, 17 October 2002. Last modified on 14 October 2003.
*/

public class SimSellke {
  /** Debug state */
  boolean DEBUG = false;
  /** Stream, where final stuff output goes */
  PrintStream out;
  /** Are we sampling sero conversions, i.e. should they appear in the log files? */
  boolean sampleSero = true;

  //Identifiers.
  public static int beta_ = 1;
  public static int betaN_ = 2;
  public static int gammaE_ = 3;
  public static int deltaE_ = 4;
  public static int gammaI_ = 5;
  public static int deltaI_ = 6;
  public static int gammaD_ = 7;
  public static int deltaD_ = 8;
  /** Data object storing the information read form the .data file */
  public static Data data;


  /** No of Units in the x direction (down) */
  int noOfUnitsX;
  /** No of Units in the x directIndivion (right->) */
  int noOfUnitsY;
  /** N vector with initial susceptibles in each unit */
  long N[][];
  /**a vector with initial susceptibles in each unit */
  long a[][];
  /** beta parameters of each unit */
  double beta[][];
  /** gammaE parameters of each unit */
  double gammaE[][];
  /** deltaE parameters of each unit */
  double deltaE[][];
  /** In case of different values than NaN we have constant length incubation time */
  double constantE[][];
  /** gammaI parameters of each unit */
  double gammaI[][];
  /** deltaI parameters of each unit */
  double deltaI[][];
  /** gammaS parameters of each unit */
  double gammaS[][];
  /** deltaS parameters of each unit */
  double deltaS[][];
  /** k parameter of each unit, i.e. \beta^{u'->u} = \kappa^{u'->u} \beta^u */
  double kappa[][][][];
  /** The time of the t\equiv 0 event, i.e. either e0 or r0.
   *  The value is expressed on the time scale, where i0=0.*/
  double origoTime = 0;
  /** Time, where epidemic definitely is dead! */
  double T = Double.NaN;
  /** Total area under the infectious trajectory */
  double aui = 0;

  /** Distribution object allowing us to sample from various densities etc. */
  Distributions prob;

  /** For formatting of doubles to an output stream. */
  static protected NumberFormat form;

  /** A list to store all events in in case we want to go back to them */
  LinkedList allEvents;

  /** Class to handle each individual in the population. */
  public class Individual implements Comparable {
    /** Location of the individual*/
    public int x,y;
    /** Individuals infection length, if infected */
    public double inflength;
    /** Individuals length of incubation period */
    public double incubtime;
    /** Time from exposure to seroconversion */
    public double serotime;

    /** The individual's threshold before becoming exposed */
    public double threshold;

    /** Comparing based on the individuals threshold
     *
     * @param o Object to compare to (of type Individual)
     * @return -1,0,1 if o1 is less than, equal to, or greater than o2
     */
    public int compareTo(Object other) {
      Individual otheri = (Individual)other;

      //Branch on threshold.
      if (this.threshold < otheri.threshold)  {return(-1);}
      if (this.threshold == otheri.threshold) {return(0);}
      if (this.threshold > otheri.threshold)  {return(1);}
      //Dummy return required by Java
      return(0);
    }

    /** Return the eventType event for the individual. Uses allEvents
     *
     * @param eventType The eventype to search for: I, R, S
     * @returns The inf. event of the individual. Null if it does not exist.
     */
    public Event findEvent(char eventType) {
      Iterator iter = allEvents.iterator();
      for (int i = 0; i < allEvents.size(); i++) {
        Event e = (Event)iter.next();
        if ((e.who == this) && (e.type == eventType)) {
          return(e);
        }
      }
      return(null);
    }

    /** Method to write all event times of an individual to the log file
     *
     * @param origoTime Time to correct all event times for.
     * @returns One line string with format: posx posy inftime rectime serotime
     */
    public String toString(double origoTime) {
      StringBuffer res = new StringBuffer();

      //Find the three events of the individual.
      Event exp  = findEvent('E');
      Event inf  = findEvent('I');
      Event sero = findEvent('S');
      Event rec  = findEvent('R');

      //Write it to the result string
      res.append(x + "\t" + y + "\t" + form.format(exp.t - origoTime) + "\t" + form.format(inf.t - origoTime) + "\t" +  form.format(rec.t - origoTime));
      //Seroconversion time.
      res.append("\t" + form.format(sero.t - origoTime));

      //Return
      return(res.toString());
    }

    /** Constructor */
    public Individual(int x, int y) {
      this.x = x;
      this.y = y;

      //Incubation time, fixed or Gamma distributed. -1 in constantE decides.
      if (Double.isNaN(constantE[x][y])) {
        incubtime = prob.gammaRand(gammaE[x][y],deltaE[x][y]);
      } else {
        incubtime = constantE[x][y];
      }

      //Sample time from exposure to seroconversion.
      serotime = prob.gammaRand(gammaS[x][y],deltaS[x][y]);

      //Sample the length of the infectious period.
      inflength = prob.gammaRand(gammaI[x][y],deltaI[x][y]);
      //inflength = prob.exponentialRand(1); //1/theta

      //Sample the threshold
      threshold = prob.exponentialRand(1);
    }
  }

  /** A class to handle events */
  public class Event implements Comparable{
    /**Location of the event, i.e. the individual */
    public Individual who;
    /** Event type*/
    public char type;
    /** Time of the event */
    double t;
    /** Comparing based on the event time
     *
     * @param o Object to compare to (of type Event)
     * @return -1,0,1 if o1 is less than, equal to, or greater than o2
     */
    public int compareTo(Object other) {
      Event otherev = (Event)other;

      //Branch on threshold.
      if (this.t < otherev.t)  {return(-1);}
      if (this.t == otherev.t) {return(0);}
      if (this.t > otherev.t)  {return(1);}
      //Dummy return required by Java
      return(0);
    }

    /** Ordinary constructur */
    public Event(Individual who, double t, char type) {
      this.who = who;
      this.type = type;
      this.t = t;
    }
    /** Just show it on the screen, pos, time (corrected for origoTime) and type */
    public String toString(double origoTime) {
      return(who.x + " " + who.y + " " + form.format(t-origoTime) + "\t" + type);
    }
    /** Standard toString nothing fancy here. */
    public String toString() {
      return toString(0);
    }
  }

  /**
   * Stupid function to write header of final size results
   */
  public void writeResultHeader() {
    out.print("T\t");
    for (int i=1; i<=this.noOfUnitsX*this.noOfUnitsY; i++) {
      out.print("u" + i + "\t");
    }
    out.println("Z\tZt90\tZt30\taui");
  }

  /** Function writes the header of the .data file based on the current
   *  parameter values. This can then be fed into the LadyBug program for estimation.
   *
   * @param data The stream containing the data file.
   */
  public void writeDataHeader(PrintStream data) {
    //Spoiler
    data.println("//---------------------------===----------------");
    Calendar cal = Calendar.getInstance(TimeZone.getDefault());
    java.text.SimpleDateFormat sdf = new java.text.SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
    data.println("// Data file automatically generated by SimSellke"); 
    data.println("// Date: " +  sdf.format(cal.getTime()));
    data.println("//---------------------------===----------------");


    //For each unit we specify location and initial sizes (and currently also parameters, but they'll be equal)
    for (int x=1; x<=noOfUnitsX; x++) {
      for (int y=1; y<=noOfUnitsY; y++) {
        data.println("unit " + x + " " + y + " " + N[x][y] + " " + a[x][y]);
      }
    }
    
    
    //Prior for beta and betan is Gamma(true value, 1).
    data.println("beta " + beta[1][1] + " 1 (" + beta[1][1] + ")");
    data.println("betan " + kappa[1][1][1][1]*beta[1][1] + " 1 (" + kappa[1][1][1][1]*beta[1][1]  + ")");
    //Prior for incubation time
    data.println("incu " + gammaE[1][1] + " 1 (" + gammaE[1][1] + ")\t" + deltaE[1][1] + " 1 (" + deltaE[1][1] + ")");
    //Prior for length of infectious period
    data.println("inf " + gammaI[1][1] + " 1 (" + gammaI[1][1] + ")\t" +  deltaI[1][1] + " 1 (" + deltaI[1][1] + ")");
    //Prior for length to diagnosis time
    data.println("diag " + gammaS[1][1] + " 1 (" + gammaS[1][1] + ")\t" + deltaS[1][1] + " 1 (" + deltaS[1][1] + ")");

    //If incubation time constant. Currently only a global thing is handled.
    if (!Double.isNaN(constantE[1][1])) { data.println("incu " + constantE[1][1]); }
    data.println("T " + form.format(T));
    data.println("//x     y       E       I       R       D");
  }

  /** Function to convert index to a position in the grid.
   *  Labelling is (1,1)=1, (2,1)=2
   *
   * @param index The index in the rid
   * @return The x coordinate
   */
  public int index2GridX(int index) {
    return(index / noOfUnitsX + 1);
  }
  /** Function to convert index to a position in the grid.
   *  Labelling is (1,1)=1, (2,1)=2
   *
   * @param index The index in the rid
   * @return The y coordinate
   */
  public int index2GridY(int index) {
    return(index % noOfUnitsX+1);
  }
  /** Function to convert (x,y) coordinate to an index
   *  Labelling is (1,1)=1, (2,1)=2
   *
   * @param x The x coordinate
   * @param y The y coordinate
   * @return The index
   */
  public int gridXY2Index(int x, int y) {
    return((y-1)*noOfUnitsX + x);
  }
  /** Determine whether two units are neighbours or not in N_4
   *
   * @param x1 x coordinate of first unit
   * @param y1 y coordinate of first unit
   * @param x2 x coordinate of second unit
   * @param y2 y coordinate of second unit
   *
   * @return boolean
  */
  public boolean areNeighbours(int x1, int y1, int x2, int y2) {
      int dx = (x1-x2);
      int dy = (y1-y2);
      //Must have that along N_4 direction and that first unit != second unit
      //This only works for two sections. Not enough that zero in one direction!
      //return( (dx*dy == 0) && !((dx==0) && (dy==0)) );
      return( (dx*dy==0) && (Math.abs(dx)<=1) && (Math.abs(dy)<=1) && !((dx==0) && (dy==0)) );
  }


  /** Show N* for each unit, i.e. number of initial susceptibles
   *  affected by the epidemic.
   *
   *  @param N Number of initial Susceptibles
   *  @param S Number of susceptibles at end (or at least currently).
   *  @return Final size in each unit as well as their sum (as a string)
   */
  public String finalSize(long N[][], TreeSet S[][]) {
    StringBuffer s = new StringBuffer();
    long sum = 0;
    for (int x=1; x<=noOfUnitsX; x++) {
      for (int y=1; y<=noOfUnitsY; y++) {
        sum += N[x][y] - S[x][y].size();
        s.append(N[x][y] - S[x][y].size() + "\t");
      }
    }
    return(s.toString() + sum);
  }

  /** Calculate, Z(t), the number of initial susceptibles
   *  affected by the epidemic (calced from current state).
   *
   *  @param N Number of initial Susceptibles
   *  @param S Number of susceptibles at end (or at least currently).
   *  @return Final size in each unit as well as their sum (as a string)
   */
  public long noOfAffected(long N[][], TreeSet S[][]) {
    StringBuffer s = new StringBuffer();
    long sum = 0;
    for (int x=1; x<=noOfUnitsX; x++) {
      for (int y=1; y<=noOfUnitsY; y++) {
        sum += N[x][y] - S[x][y].size();
      }
    }
    return(sum);
  }

  /** Show current X(t) and Y(t) for all units */
  public void state(TreeSet S[][], TreeSet E[][], TreeSet I[][], TreeSet R[][]) {
    System.out.println("------");
    for (int x=1; x<=noOfUnitsX; x++) {
      for (int y=1; y<=noOfUnitsY; y++) {
        System.out.println("Unit (" + x + "," + y + ")");
        System.out.println("X = " + S[x][y].size() + "\tY = " + I[x][y].size());
      }
    }
  }


  /** Write the corresponding events into the logfiles
   *  @param baseName baseName.rec and baseName.inf are made
   * Deprecated version with rec and inf file. Now just one .data file
   */
  public void writeLogFiles(String baseName) {
    //Write to the logFiles.
    try{
      PrintStream data = new PrintStream(new FileOutputStream(baseName + ".data",false));
      PrintStream all  = new PrintStream(new FileOutputStream(baseName + ".all",false));

      /////////////////////////////////////////////////////////////////
      //Write all events to the .all file without any further thoughts.
      /////////////////////////////////////////////////////////////////
      
      Iterator iter = allEvents.iterator();
      for (int i = 0; i < allEvents.size(); i++) {
        Event e = (Event)iter.next();
        //Stuff it into the all file
        all.println(e.toString(origoTime));
      }
      all.close();

      ///////////////////////////////////////////////////////
      // Make the .data file
      //////////////////////////////////////////////////////
      //Write the header of the .rec file
      writeDataHeader(data);

      iter = allEvents.iterator();
      for (int i = 0; i < allEvents.size(); i++) {
        Event e = (Event)iter.next();
        if (e.type == 'R') {
          //Take the individual and write it to the log file
          data.println(e.who.toString(origoTime));
        }
      }
      data.close();
    } catch (IOException e) {
      System.out.println(e);
    }
  } //end of writeLog

  /** Read the specification from the data file. We share the .data specification
   * with the LadyBug program. This works coz the parser sets the necessary variables?
   *
   * @param fileName Name of the .data file...
   */
  public static Data readDataFile(String fileName) {
    System.out.println("---------- Reading data ----------\n");
    Data data = null;
    //Open the file for reading
    try {
      //Create parser and associate file to it.
      DataFileSpec parser = new DataFileSpec(new FileInputStream(fileName));

      try {
        //Do the parsing, i.e. start with first production.
        parser.DataFile();
        data = parser.getData();
        //Make sure our data is valid.
        if (!data.isOk()) {
          System.exit(0);
        }
      } catch (ParseException e) {
        System.err.println("Error: Problems while parsing " + fileName + ".");
        System.out.println(e);
        System.exit(0);
      }
    } catch (java.io.FileNotFoundException e) {
      System.out.println(e);
      System.exit(0);
    }
    return(data);
  }


  /** Constructor - setup the model parameters and initializes the
      necessary arrays.

      @param fileName File name of .data file where the setup is read from.
      @param prob A distributions object for sampling etc.
      @param baseName The basename giving where to store the basename.rec and basename.inf files
      @param origoTime Which event indicates time=0, can be either e0 or r0.
      @param writeResult Should we write all results to log files?
      @param header A header to std out in case of final sizes?
      @param vac Vaccine effect.
      @param vaccBetween Vaccine effect only on the between spread, i.e. on kappa.
  */
  public SimSellke(Distributions prob, String baseName, PrintStream out, String origoEvent, boolean writeResult, boolean header, double vac, boolean vaccBetween) {
    //Set the sampler RN generator
    this.prob = prob;

    //Set the output stream
    this.out = out;

    //Set the global variable indicating whether we should sample seroconversions
    this.sampleSero  = sampleSero;

    //Setup the list for all events
    allEvents = new LinkedList();

    ////////////////////////////////////////////////////////
    // Init SEIR specific parameters
    ///////////////////////////////////////////////////////
    noOfUnitsX = data.noOfUnitsX;
    noOfUnitsY = data.noOfUnitsY;

    //Init N,a, the number of initial susceptible and initial infectives.
    N = new long[noOfUnitsX+1][noOfUnitsY+1];
    a = new long[noOfUnitsX+1][noOfUnitsY+1];
    for (int x=1; x<=noOfUnitsX; x++) {
      for (int y=1; y<=noOfUnitsY; y++) {
        N[x][y] = data.units[x][y].N; //15 normally
        a[x][y] = data.units[x][y].a;
      }
    }

    //Init beta, gammaI and kappa
    beta  = new double[noOfUnitsX+1][noOfUnitsY+1];
    gammaE = new double[noOfUnitsX+1][noOfUnitsY+1];
    deltaE = new double[noOfUnitsX+1][noOfUnitsY+1];
    constantE = new double[noOfUnitsX+1][noOfUnitsY+1];
    gammaI = new double[noOfUnitsX+1][noOfUnitsY+1];
    deltaI = new double[noOfUnitsX+1][noOfUnitsY+1];
    gammaS = new double[noOfUnitsX+1][noOfUnitsY+1];
    deltaS = new double[noOfUnitsX+1][noOfUnitsY+1];
    kappa = new double[noOfUnitsX+1][noOfUnitsY+1][noOfUnitsX+1][noOfUnitsY+1];
    for (int x=1; x<=noOfUnitsX; x++) {
      for (int y=1; y<=noOfUnitsY; y++) {
        //beta[x][y] depends on what type of vaccination effect there is.
        beta[x][y]  = vaccBetween ?  data.startVal[beta_] : data.startVal[beta_]*(1-vac) ;

        //Length of incubation time. In final size calcs this can be ignored.
        gammaE[x][y] = data.startVal[gammaE_];
        deltaE[x][y] = data.startVal[deltaE_];
        constantE[x][y] = Double.NaN;

        //Infection time. I ~ \Gamma(gammaI,deltaI), i.e. mean gammaI/deltaI and variance gammaI/deltaI^2
        gammaI[x][y] = data.startVal[gammaI_]; //150, //1.8200827177242969; //140 10
        deltaI[x][y] = data.startVal[deltaI_]; //10; //0.12309609779063006; //\Gamma(1,4) \equiv \Exp(4)

        //Time to seroconversion. S ~ \Gamma(gammaS,deltaS). Not important wrt. final size.
        gammaS[x][y] = data.startVal[gammaD_];
        deltaS[x][y] = data.startVal[deltaD_];

        for (int xx=1; xx<=noOfUnitsX; xx++) {
          for (int yy=1; yy<=noOfUnitsY; yy++) {
            //Effect of infectives in (xx,yy) on (x,y) relative to beta[x][y]
            kappa[xx][yy][x][y] =   vaccBetween ? data.startVal[betaN_]/data.startVal[beta_]*(1-vac) : data.startVal[betaN_]/data.startVal[beta_];
          }
        }
      }
    }

    //System.out.println("beta= " + beta[1][1]);
    //System.out.println("betaN= " + kappa[1][1][1][1]*beta[1][1]);

    //Setup up all individuals belonging to one compartment in the SEIR model
    TreeSet S[][] = new TreeSet[noOfUnitsX+1][noOfUnitsY+1];
    TreeSet E[][] = new TreeSet[noOfUnitsX+1][noOfUnitsY+1];
    TreeSet I[][] = new TreeSet[noOfUnitsX+1][noOfUnitsY+1];
    TreeSet R[][] = new TreeSet[noOfUnitsX+1][noOfUnitsY+1];

    //A queue containing events
    TreeSet evq = new TreeSet();

    for (int x=1; x<=noOfUnitsX; x++) {
      for (int y=1; y<=noOfUnitsY; y++) {
        S[x][y] = new TreeSet();
        E[x][y] = new TreeSet();
        I[x][y] = new TreeSet();
        R[x][y] = new TreeSet();

        //Create all the intially susceptibles.
        for (int i=0; i<N[x][y]; i++) {
          S[x][y].add(new Individual(x,y));
        }
        //Create all the initially infected individuals. This is done by
        //adding initially adding them to the exposed set.
        for (int i=0; i<a[x][y]; i++) {
          Individual in = new Individual(x,y);
          //Add the exposure even time occuring at time zero - i.e. all initial occur at time zero
          evq.add(new Event(in,0,'E'));
        }
      }
    }

    //Accumulated and current pressure in each unit.
    double accPressure[][] = new double[noOfUnitsX+1][noOfUnitsY+1];
    double incPressure[][] = new double[noOfUnitsX+1][noOfUnitsY+1];

    //An iteration object
    Iterator iter;
    //Absolute time -- note t==0 equals the first infection.
    double t=0;
    //Time of first recovery.
    double r0=0;
    //Have we seen any recoveries so far?
    boolean firstRecovery = true;
    //Number of affected at time t=90
    double Zt90 = 0;
    double Zt30 = 0;

    //Run the epidemic as long as there are pending events.
    while (!evq.isEmpty()) {
      //Do we need debug info?
      if (DEBUG) {
        System.out.println("=============>>\n");
        System.out.println("The event queue at time t=" + form.format(t));
        iter = evq.iterator();
        for (int i = 0; i < evq.size(); i++) {
          System.out.println((Event)iter.next());
        }
      }

      //If time is equal to 90 then we are interested in the final size, coz
      //this is what matters from a production point of view.
      if ((t<=90)) {
        long Zt = this.noOfAffected(N,S);
        if (t<=90) {
          Zt90 = Zt;
        }
        if (t<=30) {
          Zt30 = Zt;
        }
      }

      //Two options for an event exist - either it is one waiting
      //on the event queue or the total pressure is cracking someone (i.e. min of all)
      TreeSet nextEvent = new TreeSet();

      //Add the first from the event queue as a candidate
      nextEvent.add(evq.first());

      //Add all candidate events for a crack. Simulateneously
      //do what it takes to calculate \int_{0}^(t) \beta I(t) + neighbours for all units.
      for (int x=1; x<=noOfUnitsX; x++) {
        for (int y=1; y<=noOfUnitsY; y++) {
          //Calculate current pressure increase per time unit for each population.
          //But only in case there are susceptibles left.
          if (!S[x][y].isEmpty()) {
            //Pressure in the unit itself
            incPressure[x][y] = beta[x][y] * I[x][y].size();
            //Neighbour pressure.
            for (int xx=1; xx<=noOfUnitsX; xx++) {
              for (int yy=1; yy<=noOfUnitsY; yy++) {
                if (areNeighbours(x,y,xx,yy)) {
                  incPressure[x][y] += kappa[xx][yy][x][y] * beta[x][y] * I[xx][yy].size();
                }
              } //for yy
            }//for xx

            //Which individual in the unit would be the next to crack?
            Individual next =  (Individual)S[x][y].first();

            //How many time units at units' current pressure before the next cracks. If
            //no pressure we can wait forever!
            double failTime;
            if (incPressure[x][y] != 0) {
              //This assumes linear growth..is this correct? Yes, coz its a step func.
              failTime = (next.threshold-accPressure[x][y])/incPressure[x][y];
            } else {
              failTime = 1e99;
            }

            //Propose a crack event
            nextEvent.add(new Event(next,failTime + t,'E'));
          }
        }
      }


      //Act based on event type
      Event e = (Event)nextEvent.first();
      Individual who = e.who;

      //Add pressure to all units
      for (int x=1; x<=noOfUnitsX; x++) {
        for (int y=1; y<=noOfUnitsY; y++) {
          accPressure[x][y] += incPressure[x][y] * (e.t - t);

          if (DEBUG) {
            System.out.println("t=" + e.t + "\taccPressure["+x+"]["+y+"]=" + accPressure[x][y]);
            System.out.println("incPressure["+x+"]["+y+"]=" + incPressure[x][y]);
          }
        }
      }
      //All done, we are ready to increase time
      t = e.t;

      //Add the event to the list of all events
      allEvents.add(e);

      //Perform the eventtype specific actions.
      switch (e.type) {
        case 'E':
          //Move individual from S set to E set
          S[who.x][who.y].remove(who);
          E[who.x][who.y].add(who);
          //Add an infection event, when incubation time is over.
          evq.add(new Event(who,t+who.incubtime,'I'));
          //Add a seroconversion event when we are done waiting for that
          evq.add(new Event(who,t+who.serotime,'S'));
          break;
        case 'I':
          //Move individual from E set to I set
          E[who.x][who.y].remove(who);
          I[who.x][who.y].add(who);
          //Add a recovery event, when infectious period is over
          evq.add(new Event(who,t+who.inflength,'R'));
          //Add the individuals inf time to the (AUI) Total area under trajectory of infectives.
          aui += who.inflength;
          break;
        case 'R':
          //Check if its the first recovery?
          if (firstRecovery) {
            r0 = e.t;
            firstRecovery = false;
          }
          //Move individual from I set to R set
          I[who.x][who.y].remove(who);
          R[who.x][who.y].add(who);
          break;
        case 'S':
          //Nothing do to
          break;
      }
      //If the event came from the event queue then remove it. This is always
      //the case for S,R,I events, but also for the initial E events.
      if (evq.first() == e) { evq.remove(e); }

      //Now we can just look at the top of the stack to see which event it will
      //be - the rest is discareded.
      if (DEBUG) {
        System.out.println("Next event = " + (Event)nextEvent.first());
        state(S,E,I,R);
      }
    } //end of while

    /////////////////////////////////////////////////////////////
    //We are done simulating - ready to postprocess
    ////////////////////////////////////////////////////////////

    //////////////////////////////////////////////////
    //Set the time origo relative to t=0 @ i0 scale.
    //////////////////////////////////////////////////
    if (origoEvent.equals("r0")) {
      origoTime = r0;
    } else if (origoEvent.equals("e0")) {
      //if there is more than one initial infective this does not really work.
      origoTime = 0;
    } else {
      System.err.println("No such event \"" + origoEvent + "\"");
       }


    //Set time, where epidemic is ceased.
    T = ((Event)allEvents.getLast()).t+1-origoTime;

    //Write all the trajectories etc. to log files?
    if (writeResult) {
          writeLogFiles(baseName);
    }
    //If a header write out the unit names, Z and aui
    if (header) {
      writeResultHeader();
    }
    //Show the final size result
    //if ((N[1][1] - S[1][1].size()) == 0) {
    //  System.out.print(form.format(0) + "\t");
    //}
    //System.out.println(finalSize(N,S)+"\t" + form.format(aui));
    out.println(form.format(T) + "\t" + finalSize(N,S)+"\t" + Zt90 + "\t" + Zt30 +"\t"+ Math.round(100*aui)/100.0);
  }


  /** Method to initialize the sampler. Move out of
   * the instance class, coz that might mess things up and create
   * non randomness calling that too many times once for each epidemic.
   */
  static public Distributions newSampler() {
    // Seed value for Random number generator (this is a little dirty!)
    java.util.Random r = new java.util.Random();
    int seed = Math.abs(r.nextInt()); //should be 1 - 2^31-1 and notjust abs

    //Fixed seeding - in order to be able to restore results.
    //seed = 23132;

    //PRNG generating
    CollingsPRNGAdministrator a = new CollingsPRNGAdministrator(seed);
    PRNG prng = new CollingsPRNG( a.registerPRNGState() );

    //Create distribution object.
    return(new Distributions(prng));
  }

  /** Method to make the calculations for the VaccineEffect Figure shown in the Article.
   * This is a version which used the posterior predictive probability
   *
   * @param prob Link to a Distributions object useful when generating numbers
   * @param logFileName of the logfile to extract the posterior samples from.
   * @param noOfSamplesPerSample Number of samples to generate for each pair in the posterior-samples.
   * @param between Are both beta and betaN reduce (false) or just betaN (true)
   */

  static private void simVaccineEffect(Distributions prob, String logFileName, int noOfSamplesPerSample, boolean between) {
    System.out.println("Simulating the effect of a vaccination...");
    System.out.flush();
    double vac = 0;
    for (int fileCounter = 0; fileCounter <= 21; fileCounter++) {
      //Move in 0.1 steps from 0.0 to 0.6
      if (fileCounter <= 6) {
        vac = Math.round(fileCounter * 10) / 100.0;
        //From 0.6 to 0.8 in steps of 0.4
      }
      else if ( (fileCounter >= 7) && (fileCounter <= 11)) {
        vac = Math.round( (fileCounter - 6) * 4 + 60) / 100.0;
      }
      else {
        //Move in 0.2 steps starting at 0.82
        vac = Math.round( (fileCounter - 11) * 2 + 80) / 100.0;
      }

      //A multitude of catching is done by trying....
      try {
        //Open the log file
        PrintStream ps = new PrintStream(new FileOutputStream("final-" + fileCounter + ".txt"));
        System.out.println("Currrent vaccination effect: " + vac);

        //////////////////////////////////////////////////////////////////////////////////////////
        //Instead of simulating all samples from a fixed point estimate this new version uses
        //posterior samples to calculate a predictive posterior distribution
        BufferedReader buf = new BufferedReader(new FileReader(logFileName));

        //Skip header
        String aLine = buf.readLine();

        //Loop over all samples in the file - i.e. read all Lines
        int j=0;
        while ( ( (aLine = buf.readLine()) != null)) {
          //Tokenize it. There prob. is a smarter way to do this. But currently it works.
          StringTokenizer st = new StringTokenizer(aLine);

          //Read beta,betaN,gammaE,deltaE,gammaI,deltaI and throw it into the data obj.
          //This works coz the SimSellke class uses the values in data.startVal.
          //The format of the log file is loglik beta betaN gammaE deltaE gammaI deltaI gammaD deltaD
          double loglik = form.parse(st.nextToken()).doubleValue();
          for (int i = beta_; i <= deltaD_; i++) {
            data.startVal[i] = form.parse(st.nextToken()).doubleValue();
          }
          
          //Now use this posterior sample to generate <noOfSamplesPerSample>
          for (int i = 0; i < noOfSamplesPerSample; i++) {
            SimSellke mFinal = new SimSellke(prob, "foobar", ps, "r0", false, (j == 0) && (i==0)/*write header only once*/,
                                             vac,
                                             between);
          }
          j++;
        }
        //Done with all samples. Now the log file can be closed.
        //System.out.println("Generated a sampe of size: " + j + " * " + noOfSamplesPerSample);
        buf.close();
        ps.close();
      }
      catch (Exception e) {
        System.out.println(e);
      }

    }
  }

  /**
     A sir sampler for command line calling.
  */
  static public void main(String[] argv ) {
    //Argv indexes
    int _DATAFILE = 0;
    int _WHICH = 1; //type of simulation: final for finalsize and traj for trajectory
    int _ORIGO = 2; //origo e0 or r0
    int _SERO  = 3; //include seroconversions into dataset yes/no
    int _LOGFILE = 2; //Name of the estimate log file.
    int _VACCEFFECT = 3; //all or between
    int _NOOFSAMPLESPERSAMPLE = 4; //how many samples to generate for each
    //Length of a run.
    long maxruns = 10000;
    Distributions prob = newSampler();

    //Setup the formatting to US style and 5 digits
    form = NumberFormat.getInstance(Locale.ENGLISH);
    form.setMinimumFractionDigits(4);
    form.setMaximumFractionDigits(4);

    //Show splash screen
        System.out.println(
           "\n-=-=-=-=-=-=-=-=-=-=-=--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-");
        System.out.println(
           "-      \\./     -=-=  SimSellke    by Michael Hoehle              =-=-=-=-=");
        System.out.println(
           "-     (o o)    -=-=  http://www.stat.uni-muenchen.de/~hoehle/    =-=-=-=-=");
        System.out.println(
           "=-oOOo-(_)-oOOo-=-=-=-=-=--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=");
        System.out.println("-                                                                        -");
        System.out.println("- SimSellke comes with ABSOLUTELY NO WARRANTY; for details see         -");
        System.out.println("- http://www.gnu.org/copyleft/gpl.html (GPL). This is free software,     -");
        System.out.println("- and you are welcome to redistribute it under certain conditions        -");
        System.out.println("- as specified by GPL.                                                   -"); 
        System.out.println("-                                                                        -");
        System.out.println("=-=-=-=-=-=-=-=-=-=-=-=-=--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=\n");


    try {
      /** Data object storing the information read form the .data file. */
      SimSellke.data = SimSellke.readDataFile(argv[_DATAFILE]);

      //Switch on the functionality
      if (argv[_WHICH].equals("vacc")) {
        //Vaccine effect - so much going on that we stuffed it in a procedure.
        simVaccineEffect(prob,argv[_LOGFILE], Integer.parseInt(argv[_NOOFSAMPLESPERSAMPLE]), argv[_VACCEFFECT].equals("between"));
        //No vacc alternative
        //SimSellke mFinal = new SimSellke(prob,"foobar",System.out,"r0",false, i==0 && vac==0,vac);
      }  else if (argv[_WHICH].equals("final")) {
          try {
            System.out.println("Simulating 10000 final sizes. Monitor final.txt for progress...");
            System.out.println("Temporary files foobar.all and foobar.data are written.");
            PrintStream ps = new PrintStream(new FileOutputStream("final.txt"));
            for (int i=0; i<10000; i++) {
              SimSellke mFinal = new SimSellke(prob,"foobar",ps,"r0",false, i==0 /*write header for each file*/,0,false);
            }
            ps.close();
            System.out.println("Done.");
          } catch (FileNotFoundException e) {
          }
      } else if (argv[_WHICH].equals("traj")) {
        System.out.println("Simulating an epidemic. See foobar.all and foobar.data");
        SimSellke mTraj = new SimSellke(prob,"foobar",System.out,argv[_ORIGO],true,false,0,false);
      } else { //illegal option.
        throw(new ArrayIndexOutOfBoundsException());
      }
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Syntax Error:\n\n" +
                         "simSellke <DATAFILE> ( vacc <LOGFILE> all|between <noOfSamplesPerSample>\n"  +
                         "                     | final\n" +
                         "                     | traj e0|r0).\n");
    }
  }
}

