package sir.util;

import java.util.Vector;
import java.util.Hashtable;
import sir.estimate.*;

/**
 * Title:        LadyBug 2.0
 * Description:  This is Object handles the data reading (i.e. the .data) files.
 * It contains all the data which are then moved to an instance of the EpiState class.
 * Copyright:    Copyright (c) 2003
 * Company:
 * @author Michael Hhle <hoehle@stat.uni-muenchen.de>
 * @version 2.0
 */

public class Data {
    //This results in a high connectivity between model and data..
    String modelParamNames[] = new String[] {"FORTRAN","beta","betaN","gammaE","deltaE","gammaI","deltaI","gammaD","deltaD"};
    int noOfModelParams = 8;
    //Hashtable containing the prior distribution values.
    public double nu[] = new double[noOfModelParams+1];
    public double lambda[] = new double[noOfModelParams+1];
    public double startVal[] = new double[noOfModelParams+1];

    /** Exposure times   */
    public Event E[] = null;
    /** Infectious times */
    public Event I[] = null;
    /** Recovery times   */
    public Event R[] = null;
    /** Diagnostic Events*/
    public Event D[] = null;
    /** End of observation*/
    public Event T = null;

    /* 2D Array with units */
    public Unit[][] units;

    /** NullUnit, special void unit to use in case of switch and terminating events */
    public Unit nullUnit = new Unit(-1,-1,0,0,false);
    /** Pointer to the unit, where the first infection occurs (i.e. a=1) */
    public Unit firstInfectedUnit;
    /** Index of the inoculated individuals */
    public int inoculatedIndex;

    /** Various sorts of counters */
    public int noOfUnits =0, noOfUnitsX =0, noOfUnitsY =0;
    /** Number of affected individuals, i.e. number of individuals mentioned in the data file */
    public int noOfIndividuals =0;
    public int noOfBetaN=0;
    public int noOfBeta=0;

    /** Length of incubatoin period */
    public double incu = Double.NaN;

    /** Time, where experiment has ceased. */
    public double Ttime = Double.NaN;
    /** Exposure of first exposed */
    public Event E1 = null;

    /** Function to check that the data found in the data file is ok.
     *  I.e. switch time order and recov/inf pairs have to be ok.
     *
     * @return A boolean indicitating whether data is good.
     */
    public boolean isOk() {
      boolean res = true;
      //Check that data times are ok.
      for (int i=0; i<noOfIndividuals; i++) {
        if (I[i].isParametric || R[i].isParametric) {
          if (I[i].isParametric && R[i].isParametric) {
            System.err.println("Error: (Individual " + i + ") Can't have both rec=NA and inf=NA.");
            res = false;
          }
        } else if (R[i].t < I[i].t) {
          System.err.println("Error: (Individual " + i + ") Inf has to occur before rec.");
          res = false;
        }
      }

      //Return result.
      return res;
    }

    /** Check, that all cells in the grid layout obtained from the units are specified
     *  Convert units vector to a 2D array, and setup up unit related parameters
     *
     *  @param unitsVector All units just stuffed into a Vector
     */
    public boolean setupUnitArrays(Vector unitsVector) {
      //Find the dimension of the unit grid layout
      int maxx=0, maxy=0;
      //Find unit with the first infection
      firstInfectedUnit = null;
      //Both dimension finding and first infection finding happens in this loop
      for (int i=0; i<noOfUnits; i++) {
        Unit u = (Unit)unitsVector.elementAt(i);
        System.out.println("Looking at unit [" + u.posx + "][" + u.posy + "]");
        //Dimension deducing
        if (maxx<u.posx) { maxx = u.posx;}
        if (maxy<u.posy) { maxy = u.posy;}
        //Find unit with first infection
        if (u.a>0) {
          if ((u.a==1) && (firstInfectedUnit==null)) {
            firstInfectedUnit = u;
            //System.out.println("First infection is in unit (" + u.posx + "," +  u.posy + ")");
          } else {
            System.err.println("Error: [Unit " + u.posx + " " + u.posy + "] More than one initial infective.");
            System.exit(0);
          }
        }
      }
      //If no initial infective - we got a problem.
      if (firstInfectedUnit == null) {
        System.err.println("Error: No initial infective!");
        System.exit(0);
      }

      noOfUnitsX = maxx;
      noOfUnitsY = maxy;
      System.out.println("Unit layout is " + noOfUnitsX + " x " + noOfUnitsY);

      //New a new unit 2Dim array and assign all present units.
      units = new Unit[noOfUnitsX+1][noOfUnitsY+1];
      for (int i=0; i<noOfUnits; i++) {
        Unit u = (Unit)unitsVector.elementAt(i);
        //Make the units array point to the right spot.
        units[u.posx][u.posy] = u;
      }

      //Now check if there were some units missing, i.e.
      //make sure all tiles in the layout are specified, i.e. no grid-cell unspecified
      boolean ok = true;
      for (int i=1; i<=maxx; i++) {
        for (int j=1; j<=maxy; j++) {
          if (units[i][j] == null) {
            ok = false;
            System.err.println("Error: Unit (" + i + "," + j + ") is unspecified.");
          } else {
            //Fetch all local recovery times. - are we actually using that?
            units[i][j].fetchLocalRec(R,noOfIndividuals);
          }
        }
      }
      return(ok);
    }
    /** Extract local recovery times for each unit and deduce the number of individuals
     *  escaping infection in all units. */
    public void extractLocalTau() {
      for (int i=1; i<=noOfUnitsX; i++) {
        for (int j=1; j<=noOfUnitsY; j++) {
          units[i][j].fetchLocalRec(R,noOfIndividuals);
        }
      }
    }
    /** Find index of all unknown exposure time
     * @return A new'ed array with the index of all unknown exposure times @see getNoOfUnknownE
     */
    public int[] getUnknownEIndexes() {
      int res[] = new int[getNoOfUnknownE()];
      int index = 0;
      for (int i=0; i<noOfIndividuals; i++) {
        if (E[i].isParametric) {
          res[index++] = i;
        }
      }
      return(res);
    }

    /** Find index of all unknown diagnoed time
     * @return A new'ed array with the index of all unknown exposure times @see getNoOfUnknownS
     */
    public int[] getUnknownDIndexes() {
      int res[] = new int[getNoOfUnknownD()];
      int index = 0;
      for (int i=0; i<noOfIndividuals; i++) {
        if (D[i].isParametric) {
          res[index++] = i;
        }
      }
      return(res);
    }
    /** Find index of all unknown infection time
     * @return A new'ed array with the index of all unknown infection times @see getNoOfUnknownI
     */
    public int[] getUnknownIIndexes() {
      int res[] = new int[getNoOfUnknownI()];
      int index = 0;
      for (int i=0; i<noOfIndividuals; i++) {
        if (I[i].isParametric) {
          res[index++] = i;
        }
      }
      return(res);
    }
    /** Find index of all unknown recovery times
     * @return A new'ed array with the index of all unknown recovery times @see getNoOfUnknownR
     */
    public int[] getUnknownRIndexes() {
      int res[] = new int[getNoOfUnknownR()];
      int index = 0;
      for (int i=0; i<noOfIndividuals; i++) {
        if (R[i].isParametric) {
          res[index++] = i;
        }
      }
      return(res);
    }
    /** Get the number of unknown exposure times by counting number of NA's in the data
     * @return Number of unknown exposure times.
     */
    public int getNoOfUnknownE() {
      int sum=0;
      for (int i=0; i<noOfIndividuals; i++) {
        if (E[i].isParametric) { sum++; }
      }
      return(sum);
    }

    /** Get the number of unknown infection times by counting number of NA's in the data
     * @return Number of unknown infection times.
     */
    public int getNoOfUnknownI() {
      int sum=0;
      for (int i=0; i<noOfIndividuals; i++) {
        if (I[i].isParametric) { sum++; }
      }
      return(sum);
    }
    /** Get the number of unknown recovery times by counting number of NA's in the data
     * @return Number of unknown recovery times.
     */
    public int getNoOfUnknownR() {
      int sum=0;
      for (int i=0; i<noOfIndividuals; i++) {
        if (R[i].isParametric) { sum++; }
      }
      return(sum);
    }
    /** Get the number of unknown diagnosed times by counting number of NA's in the data
     * @return Number of censored diagnosed times.
     */
    public int getNoOfCensoredD() {
      int sum=0;
      for (int i=0; i<noOfIndividuals; i++) {
        if (D[i].isCensored) { sum++; }
      }
      return(sum);
    }

    /** Get the number of unknown diagnosed times by counting number of NA's in the data
     * @return Number of unknown diagnosed times.
     */
    public int getNoOfUnknownD() {
      int sum=0;
      for (int i=0; i<noOfIndividuals; i++) {
        if (D[i].isParametric) { sum++; }
      }
      return(sum);
    }
    /** Look in the data and find the last specified rec or diag time
     *
     * @return Last specified rec or diag event time specified in the data file
     */
    public double lastRecDiagTime() {
      //The last specified event time
      double lastRD = -1e99;

      //Loop over all recoveries
      for (int j=0; j<noOfIndividuals; j++) {
        if (!R[j].isParametric) {
          if (R[j].t>lastRD) { lastRD = R[j].t; }
        }
        if (!D[j].isParametric) {
          if (D[j].t>lastRD) { lastRD = D[j].t; }
        }
      }
      //Return the event time of the last evemt to,e
      return(lastRD);
    }

    /** Postprocess data, i.e. make data parametric if requested, setup units, check
     * consistency, etc. This function handles all the neceassary logic to ensure
     * the data file specified valid information.
     */
    public void postProcess() {
      //Extract local recovery events for each unit
      extractLocalTau();

      //We can only have a exponential distriution if we are NOT using a mean
      //variance representation.
      if ((Options.expInfTime||(Options.diagTime=='E')||(Options.incuTime=='E')) && Options.meanVar) {
        System.err.println("Error: Can not handle a E/V specification of a period, while Period~Exp()");
        System.exit(-1);
      }
      
      //Show fixed incubation time in case it exists   
      if (Options.readFromFile) {
        if ( (Options.incuTime == 'C') && (!Double.isNaN(incu))) {
          //as is?
          if (!Double.isInfinite(incu)) {
            System.out.println("Fixed incubation time = " + this.incu);
          }
          else {
            if (this.getNoOfUnknownE() == 0) {
              System.out.println(
                  "Incubation time is left as specified in the file.");
            }
            else {
              System.err.println(
                  "Error: \"as is\" only works if no E are parametric.");
              System.exit( -1);
            }
          }
        }
        else {
          if (! (Options.incuTime == 'E' || Options.incuTime == 'G')) {
            System.err.println("Options.incuTime = " + Options.incuTime +
                               "\tincu   = " + this.incu);
            System.err.println(
                "Error: Mismatch between incubation time specification in .sir and .data file.");
            System.exit( -1);
          }
        }
      }
      

      //Set up if I events are parametric even though they are given in data file
      if (Options.ignoreI) {
        //We are told to forget all values found in the data file
        for (int i=0; i<noOfIndividuals; i++) {
          //If it is not already parametric
          if (!I[i].isParametric) {
            //Remove all information and make it parametric.
            I[i].startT = I[i].t;
            I[i].t = Double.NaN;
            I[i].isParametric = true;
          }
        }
      }
      //Remove all information about diagnostic tests if wanted
      if (Options.ignoreD) {
        for (int i=0; i<noOfIndividuals; i++) {
          if (!D[i].isParametric) {
            D[i].startT = D[i].t;
            D[i].t = Double.NaN;
            D[i].isParametric = true;
          }
        }
      }

      //If it is censored we save the observed value (censoring time) and sample an
      //instance from the left truncated gamma distribution as stated in the article
      //Treat it as parametric from now on. Queries for censoring time should use
      //S[].censorTime
      for (int i=0; i<noOfIndividuals; i++) {
        if ((D[i].isCensored) && (!D[i].isParametric)) {
          D[i].isParametric = true;
        }
      }

      //Find first exposure
      inoculatedIndex = -1;
      for (int i=0; i<noOfIndividuals; i++) {
        if (Math.abs(E[i].t) < 0.000001) {
          E1 = E[i];
          inoculatedIndex = i;
          E[i].isParametric = false;
        }
      }

      //Remove all information about exposured if wanted. Except of course
      //e0.
      if (Options.ignoreE) {
        for (int i=0; i<noOfIndividuals; i++) {
          //If its not the first and not already parametric
          if ((E[i] != E1) && (!E[i].isParametric)) {
            E[i].startT = E[i].t;
            E[i].t = Double.NaN;
            E[i].isParametric = true;
          }
        }
      }


      //Check that if there are some recovery times missing then the data is
      //classified as a transmission experiment.
      if ((getNoOfUnknownR()>0) && (Double.isNaN(Ttime))) {
        System.err.println("Error: Missing recovery times, but the cease time T is not specified.");
        System.exit(-1);
      }


      //Find end of observation time. If T is manually specified it has to be
      //greater than the last recovery/inf time specified!
      double Tt[] = new double[2];
      Tt[0] = Double.NaN;
      if (!Double.isNaN(Ttime)) {
        Tt[1] = Ttime;
        T = new Event(units[1][1],Tt,'T');
      } else {
        Tt[1] = lastRecDiagTime()+1;
        T = new Event(units[1][1],Tt,'T');
      }
      //Check that last Time is ok!
      if (T.t<lastRecDiagTime()) {
        System.err.println("Error: T has to be greater than the last observed recovery");
        System.exit(-1);
      }
      System.out.println("T.t=" + T.t);


      //No need to check that unit information corresponds to observations, because we list
      //in a 1:1 format now.

      //Set some local unit references
      for (int i=1; i<=noOfUnitsX; i++) {
        for (int j=1; j<=noOfUnitsY; j++) {
          units[i][j].setLocalLinks(this,LadyBug.prob);
        }
      }

      // //Show information
      System.out.println("Read " + this.noOfIndividuals + " affected individuals. ");
      System.out.println("No. of unknown exposures/infections/recoveries = " + getNoOfUnknownE() + "/" + getNoOfUnknownI() + "/" + "/" + getNoOfUnknownR());
      System.out.println("No. of censored/unknown diagnostic tests       = " + getNoOfCensoredD());
    }

    /** Usual debug method to display all on screen */
    public String toString() {
      StringBuffer s = new StringBuffer();
      s.append("------------------------");
      s.append("Parameter priors:\n");
      for (int i=1;i<=noOfModelParams;i++) {
        s.append(modelParamNames[i] + "=\tGa(" + LadyBug.form.format(nu[i]) + ", " + LadyBug.form.format(lambda[i])  +")\n");
      }
      s.append("------------------------");

      //Show results.
      s.append("\nInfection, Diagnosed, and Recovery Times for each individual:\n");
      for (int i=0; i<noOfIndividuals; i++) {
        s.append(i + ":" + I[i].posx + "\t" + I[i].posy + "\t" + I[i].t + "\t" + D[i].t + "\t" + R[i].t + "\n");
      }
      for (int i=1; i<=noOfUnitsX; i++) {
        for (int j=1; j<=noOfUnitsY; j++) {
          s.append(units[i][j] + "\n");
        }
      }
      s.append("------------------------\n");
      s.append("T = " + T.t + "\n");
      s.append("------------------------\n");

      return(s.toString());
    }

    /** Function to convert index to a X position in the grid .
     *  Labelling is (1,1)=1, (2,1)=2
     *
     * @param index The index in the grid
     * @return X position
     */
    public int index2GridX(int index) {
      return(index / noOfUnitsY + 1);
    }
    /** Function to convert a index to a Y grid position
     *  Labelling is (1,1)=1, (2,1)=2
     *
     * @param index The index in the rid
     * @return The y position
     */
    public int index2GridY(int index) {
      return(index % noOfUnitsY+1);
    }

}