package sir.util;

//Import stuff
import org.omegahat.Probability.Distributions.*;
import org.omegahat.Simulation.RandomGenerators.PRNG;

/**
   Class extending the set of distributions. In the Hydra package all
   distributions are in one file not OO-split. Bad! Should use the
   CERN package for this!<P>

   Current extension consists of the gamma distribution with both
   scale and shape parameter. And incomplete gamma function.

   @see org.omegahat.Probability.Distributions.PRNGDistributionFunctions

   @author Michael Hhle &lt;<a href="mailto://hoehle@dina.kvl.dk">hoehle@dina.kvl.dk</a>&gt;
   @version 20 August 2001. Last modified 7 January 2003
*/

public class Distributions extends DistributionFunctions {
    /** The random generator */
    PRNG prng;


    /** Straight accessor function */
    public PRNG getPRNG() {
	return prng;
    }

    public void setPRNG( PRNG prng_in ) {
	prng = prng_in;
    }


    /** Constructor */
    public Distributions(PRNG prng_in) {
	prng = prng_in;
    }

    /** Overwrite the uniformRand method so we use the better PRNG!
     */
    public double uniformRand() {
	return prng.nextDouble();
    }

    /**
       Sample X ~ Gamma(alpha,beta) , s.t. E[X] = alpha/beta and
       V[X] = alpha/beta^2.
    */
    public double gammaRand (double alpha, double beta) {
	return(1/beta*gammaRand(alpha));
    }

    /** Find pdf value at X=x, when X~\Gamma(alpha,beta), i.e. E[X] = alpha/beta and
       V[X] = alpha/beta^2.
    */
    public double gammaPDF(double x,double alpha, double beta) {
        return(beta*gammaPDF(x*beta,alpha));
    }

    /** My own distribution function on log scale to make sure I got things right.
     *  Problem: This function is called heavily and should really be optimized.
     *  I don't get why StrictMath.log is called, but we have a bit too many Log's and stuff?
     * So now it's renamed to loggammaPDF2 and we have a native version. In IBM JVM it
     * appears to run better so no need for native calls there...
     */

    public double loggammaPDF(double x,double alpha, double beta) {
      //In case x<= 0 log(0) (infinite);
      if (x <= 0.0) return Math.log(0);
      if (alpha <0) return Double.NaN;
      if (beta  <0) return Double.NaN;
      //For proper x values
      return( alpha*Math.log(beta) + (alpha-1)* Math.log(x) - beta*x - lngamma(alpha));
    }

    //Native add-on - doesn't really give much performance gain despite 30% of the
    //time is spent in the StrictMAth.log function
    /*
    public native double loggammaPDF(double x, double alpha, double beta);
    static {
      System.loadLibrary("distributions");
    }
    */

    /** Constant to be used in function gammaIncomplete. Max number of iterations. */
    static long ITMAX = 10000;
    /** Tolerance to be used in function gammaIncomplete */
    static double EPS = 3.0e-7;



    /**
     * Survival function S(t) of the gamma distribution with parameters a and b
     *
     * @param x Time value
     * @param alpha Shape parameter of gamma distribution (aka. gamma)
     * @param beta 1/Scale parameter of gamma distribution.
     */
    public double gammaSurv(double x, double alpha, double beta) {
      if ((x < 0.0  )  || (Double.isNaN(x)))     return Double.NaN;
      if ((alpha <= 0) || (Double.isNaN(alpha))) return Double.NaN;
      if ((beta  <= 0) || (Double.isNaN(beta)))  return Double.NaN;
      //Use equations from Numerical Recipies, P(a,x) = cumultative prob. function.
      //and S(t) just 1-F(t), i.e. S(t) = 1-P(a,t).
      //So all we need to do is to correct for the b parameters.
      //return(1-gammaIncomplete(alpha,x*beta));
      return(1-this.gammaCDF(alpha,x*beta));
    }

    /** Quantile of a Ga(p,alpha,beta) distribution (i.e. inverse of CDF)
     *
     * @param p The prob to compute quantile for
     * @param alpha alpha param
     * @param beta beta param
     * @return qgamma(p,alpha,beta) (R syntax)
     */
    public double gammaQuantile(double p,double alpha,double beta) {
      return(1/beta*gammaQuantile(alpha,p));
    }

    /** CDF i.e. probability function of a Ga(alpha,beta) distribution
     * @param quantile The quantile
     * @param alpha alpha paramter of Ga(alpha,beta) distrib.
     * @param beta bets paramter of Ga(alpha,beta) distrib.
     * @return pgamma(q,alpha,beta)
     */
    public double gammaCDF(double q,double alpha,double beta) {
      return(gammaCDF(alpha,q*beta));
    }

    /** Sample a random number of a left truncated Ga(alpha,beta) distribution
     * using inversion sampling.
     *
     * @param alpha alpha paramter of Ga(alpha,beta) distrib.
     * @param beta bets paramter of Ga(alpha,beta) distrib.
     * @param T point of (left) truncation
     */
    public double gammaTruncRand(double alpha, double beta,double T) {
      //Find out how much it is truncated, i.e. find quantile of truncation point
      double qT = gammaCDF(T,alpha,beta);

      //Sample a uniform number in the range [pT,1] (instead of 0,1), i.e.
      //fetch a quantile
      double q = uniformRand()*(1-qT) + qT;

      //Translate gamma quantiles to x values
      return(gammaQuantile(q,alpha,beta));
    }
}
