package sir.util;

import java.util.Comparator;
import java.text.NumberFormat;
import java.util.Locale;
import sir.estimate.LadyBug;

/**
 * Title:        SIREstimate
 * Description:  Estimation of SIR parameters
 * Copyright:    Copyright (c) 2002
 * Company:
 * @author Michael Hhle <hoehle@dina.kvl.dk>
 * @version 1.0
 *
 * Added Stuff in order to compare two events to each other.
 */

 public class Event implements Comparator {
    /** Link to unit, corresponds to x,y position */
    public Unit unit;
    /** x position in grid of the unit (quick-access) */
    public int posx;
    /** y position in grid of the unit (quick-access) */
    public int posy;

    /** Time of event */
    public double t;
    /** Even type: S,E,I,R,T */
    public char type;

    /** Is the event time parametric, i.e. unknown */
    public boolean isParametric;
    /** Start value if unknown and we don't want to sample blindly we can provide a start value */
    public double startT;
    /** Acceptance number, to be used when updating by MCMC */
    public double accepted;

    /** Is the event censored. Currently - only applies to seroconversion events. */
    public boolean isCensored = false;
    /** In case an event is censored save the observed time, i.e. censoring time */
    public double censorTime = -1;

    /////////////////////////////////////////////////////////////////////
    // Modified by SirEstimate
    /////////////////////////////////////////////////////////////////////

    /** Number of susceptibles in unit posx,posy at time t*/
    public double X;
    /** Number of infected in unit posx,posy at time t*/
    public double Y;
    /** Number of exposed in unit posy,posy at time t*/
    public double Z;

    /** Link to previous event occured in same unit as this event */
    public Event prevEventInSameUnit = null;
    /** Link to the next event occured in same unit as this event */
    public Event nextEventInSameUnit = null;
    /** Next event for same individual, i.e. for E it is I, for I it is R. For R it is null */
    public Event nextIndividualEvent = null;

    // Snapshot, last event occured in all the neighbouring section.
    //  Would be nice to have but currently not supported.
    //public Event[][] unitSnapshot;

    /** The unit, where the infection is originating from. This is only relevant for infectious events..maybe we should do this by inheritance */
    //public Unit Theta; //obsolete.

    /** The unit, where the infection is originating from. This is only relevant for infectious/exposed events. */
    public Unit origin;

    //////////////////////////////////////////////////////////////////////

    /** Make an exact copy of the event, but change the event to type 'E' and subtract
     *  a constant from time. Used to find exposures from current infections. Also the
     *  origin of the exposure is samples uniformly among the set of possibilities
     *
     *  @param c The fixed length of the incubation time
     *  @return A cloned exposure event occuring at time I.t-c
     */
    public Event cloneExposed(double c) {
      double tdummy[] = new double[2];
     tdummy[0] = Double.NaN;
     tdummy[1] = t-c;
      Event e = new Event(unit,tdummy,'E');
      e.posx = posx;
      e.posy = posy;

      //Exposed events need to have an origin. To keep things valid at start
      //just always use within unit. Later on in a postprocess phase we swap to random
      e.origin = e.unit; //e.unit.randomN4Star();

      return(e);
    }

    /** Clone an event - note array instances are new'ed so the direct mem references
     * are not exactly as before
     */
    public Object clone() {
      double tdummy[] = new double[2];
      tdummy[0] = Double.NaN;
      tdummy[1] = t;
      //This covers unit, censorTime, time and type
      Event e = new Event(unit,tdummy,type);
      
      //isPArametric and isCensored
      e.isParametric = isParametric;
      e.isCensored = isCensored;
      
      //Position
      e.posx = posx;
      e.posy = posy;

      //Origin
      e.origin = origin; 
      
      //X,Y,Z
      e.X = X; e.Y=Y; e.Z=Z;
      
      //Next and previous events of various types are not clone so this is done manually
      //
      
      //Done 
      return(e);
    }
    

    /** Basic constructor transfering values into the object. To be called by more
     * specialized versions of the constructore as well.
     *
     * @param unit       Unit the event occured in
     * @param t          Time of event occurence (two dim array: place 0= startval 1=value)
     * @param type       What was the event type
     */
    public Event(Unit unit, double[] t, char type) {
        this.unit = unit;
        posx = unit.posx;
        posy = unit.posy;

	this.t = t[1];
        this.startT = t[0];

        this.type = type;

        //Exposed events need to have an origin. To keep things valid at start
        //just always use within unit. Later on in a postprocess phase we swap to random
        this.origin = unit;

        //Currently check for whether something is parametric depends on whether it is NA.
        isParametric = Double.isNaN(t[1]);
        isCensored = false;
    }

    /** Constructor with censoring
     *
     * @param unit       Unit the event occured in
     * @param t          Time of event occurence (two dim array: place 0= startval 1=value)
     * @param type       What was the event type
     * @param isCensored Censoring status, if true then t is the censoring time
     */
    public Event(Unit unit, double[] t, char type, boolean isCensored) {
      //Call the basic constructor
      this(unit,t,type);
      //Add censoring information, i.e. censored at time t
      this.isCensored = isCensored;
      this.censorTime = t[1];
    }

    /** Empty constructor if we ever need it. Just initialize the booleans. */
    public Event() {
      isParametric = true;
      isCensored = false;
    }

    /** For debug purposes */
    public String toString() {
       switch (type) {
          case 'E':
            return(LadyBug.form.format(t) + "\t" + unit.posx + " " + unit.posy + " " + type + "\t" + Z + " " + X + " " + Y + "\tO=(" + origin.posx + "," + origin.posy + ")");
          case 'S':
            String s;
            if (isCensored) {s=" (Censored)";} else {s="";}
            return(LadyBug.form.format(t) + "\t" + unit.posx + " " + unit.posy + " " + type + "\t" + Z + " " + X + " " + Y + s);
          default:
            return(LadyBug.form.format(t) + "\t" + unit.posx + " " + unit.posy + " " + type + "\t" + Z + " " + X + " " + Y);
        }
    }


    /** Comparing based on event time. Note that if time is equal we induce an order
     * which prefers E over R events in time.
     *
     * @param o1 First object (should be of type Event)
     * @param o2 Second object (also of type Event)
     * @return -1,0,1 if o1 is less than, equal to, or greater than o2
     */
     public int compare(Object o1, Object o2) {
      Event e1 = (Event)o1;
      Event e2 = (Event)o2;
      //Branch on time.
      if (e1.t < e2.t) {return(-1);}

      //If time is equal, then we will induce an order, s.t. E comes before R
      if (e1.t == e2.t) {
        if (e1.type == 'R') return(1);
        if (e2.type == 'R') return(-1);
        return(0);
      }
      if (e1.t > e2.t) {return(1);}
      //Dummy return required by Java
      return(0);
     }
}
