package sir.util;

import java.util.*;


/**
 * Title:        LadyBug
 * Description:  Estimation of SIR parameters
 * Copyright:    Copyright (c) 2002
 * Company:
 * @author Michael Hhle <hoehle@dina.kvl.dk>
 * @version 1.0
 *
 * Class holding static information for each unit
 */

public class Unit {
    /** Initial number of susceptibles */
    public int N;
    /** Initial number of infected */
    public int a;
    /** Number of individuals not affected by the epidemic at time T. I.e. number of unit individuals being part of S_T */
    public int STmembers;

    /** X-position (i.e. row number) of unit in grid */
    public int posx;
    /** Y-position (i.e. column number) of unit in grid */
    public int posy;

    /** Are animals in this unit vaccinated? */
    public boolean vacc;

    /** Link to all recovery events occuring in this unit */
    public Vector rec = new Vector();
    /** First event occuring in the unit - usually a 'N' event at time of I_0 */
    public Event firstEvent = null;

    /** Link to data object so it knows a bit about where it is located */
    protected Data data;
    /** Link to a random generator so we got one if we need it*/
    protected Distributions prob;

    /** Neighbours - precomputed so we don't need to calculate it every time */
    public LinkedList N4Star = null;
    public LinkedList N4 = null;

    public void setLocalLinks(Data data, Distributions prob) {
      this.data = data;
      this.prob = prob;
    }

    /** Constructor
     * @param posx x position in unit grid
     * @param posy y position in unit grid
     * @param N number of initial susceptible in unit
     * @param a number of initial infectives in unit
     * @param vacc Boolean indicating whether all animals in unit are vacicnated
    */
    public Unit(int posx, int posy, int N, int a, boolean vacc) {
        this.posx = posx;
        this.posy = posy;
        this.N = N;
        this.a = a;
        this.vacc = vacc;
    }
    /** Debug Method */
    public String toString() {
        StringBuffer s = new StringBuffer();
        s.append("-------- Unit (" + posx + "," + posy + ") --------------------\n");
        s.append("(N,a,vacc) = (" + N    + ", " + a + ", " + vacc + ")\n");
        s.append("------------------------------------------\n");
        return(s.toString());
    }

    /** Fetch link to all recovery events associated to this unit and once this is
     *  found compute the number of individuals from section escaping infection. I.e.
     *  N_u - (|\r_u|-a_u)
     *
     * @param globalTau Array with all recovery events.
     * @param noOfGlobalTau Number of elements in globalTau
     */
    public void fetchLocalRec(Event[] globalRec,int noOfGlobalRec) {
      int localCounter = 0;
      for (int i=0; i<noOfGlobalRec; i++) {
        if (globalRec[i].unit == this) {
          rec.add(localCounter++,globalRec[i]);
        }
      }
      //At this point we know the number of individuals from the unit
      //who escaped the epidemic
      STmembers = N - (rec.size() - a);
    }
    /** Determine whether a unit is a neighbour or not.
     *  Currently everything is a neighbour
     *
     * @return boolean
     */
    public boolean isNeighbour(Unit u) {
      int dx = (posx-u.posx);
      int dy = (posy-u.posy);
      //Must have that along N_4 direction and that first unit != second unit
      return( (dx*dy == 0) && !((dx==0) && (dy==0)) );
    }

    /** Return set of all units belonging to N4Star */
    public LinkedList N4Star() {
      //If we already computed it once.
      if (N4Star != null) {return(N4Star);}

      LinkedList res = new LinkedList();
      //Loop over all compass indeces, i.e. *, N, S, E, W
      for (int ci=0; ci<5; ci++) {
        int dx = 0;
        int dy = 0;
        switch (ci) {
          case 0: //The unit itself
            break;
          case 1: //East direction
            dx = -1;
            break;
          case 2: //North
            dy = -1;
            break;
          case 3: //West
            dx = 1;
            break;
          case 4: //South
            dy = 1;
            break;
        } //end of switch
        try {
          //Try to access the unit. If it works we are done
          Unit um = data.units[posx + dx][posy + dy];
          //Avoid 0,0 indexes - these exist, coz we start the arrays @ 1.
          if ((posx+dx>0) && (posy+dy>0)) res.add(um);
        } catch (ArrayIndexOutOfBoundsException e) {
          //Unit does not exist
        }
      } //end of for

      //Store the result
      N4Star = res;

      //Return.
      return(res);
    } //end of method.

    /** Return the compass direction neighbours
     *  @return N_4(u)
     */
    public LinkedList N4() {
      //If we already computed it once.
      if (N4 != null) {return(N4);}
      //Fetch N_4^*(u)
      LinkedList res = new LinkedList();
      Iterator it = N4Star().iterator();
      //We need to clone the list, coz the N4Star instance is used elsewhere.
      while (it.hasNext()) {
        Unit u = (Unit)it.next();
        if (u!=this) { res.add(u); }
      }
      //Store it so we don't have to compute it again.
      N4 = res;
      return(res);
    }

    /** Sample a unit from N_4^*(u) uniformly
     *
     *   @return u\in N_4^*(this)
     */
    public Unit randomN4Star() {
      Unit res = null;
      //Sample a unit from the four compass directions plus u until
      //we find one actually being valid.
      boolean validUnit = false;
      while (!validUnit) {
        //Sample an int 0,1,2,3,4
        int index = (int)Math.floor(prob.uniformRand()*5) % 5;
        int dx = 0;
        int dy = 0;
        switch (index) {
          case 0: //The unit itself
            break;
          case 1: //East direction
            dx = -1;
            break;
          case 2: //North
            dy = -1;
            break;
          case 3: //West
            dx = 1;
            break;
          case 4: //South
            dy = 1;
            break;
        }
        try {
          //Try to access the unit. If it works we are done
          res = data.units[posx + dx][posy + dy];
          //Coz we start the array at one zeros are not allowed but don't crash.
          validUnit = ((posx+dx) > 0) && ((posy+dy) > 0);
        } catch (ArrayIndexOutOfBoundsException e) {
          //Nothing to do
        }
      } //end of while
      //Return.
      return(res);
    } //end of sample
}
